# BEGIN LICENSE
# Copyright 2020, Manuel Genovés <manuel.genoves@gmail.com>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Gio, Handy

from .settings import Settings
from apostrophe.helpers import get_media_path

class Theme:
    '''Abstracts getters for the current theme related
    resources'''

    settings = Settings.new()

    def __init__(self, name, gtk_css, gtk_css_hc, web_css):
        self.name = name
        self.gtk_css    = gtk_css
        self.gtk_css_hc = gtk_css_hc
        self.web_css    = web_css

    @classmethod
    def get_for_name(cls, name, default=None):
        current_theme = default or defaultThemes[0]
        for theme in defaultThemes:
            if name == theme.name:
                current_theme = theme
        return current_theme

    @classmethod
    def get_current(cls):
        color_scheme = cls.settings.get_string("color-scheme")
        return cls.get_for_name(color_scheme)

defaultThemes = [
    Theme('light', Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita.css"),
          Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita-hc.css"),
          get_media_path('/media/css/web/adwaita.css')),
    Theme('dark', Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita-dark.css"),
          Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita-dark-hc.css"),
          get_media_path('/media/css/web/adwaita.css')),
    Theme('sepia', Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita-sepia.css"),
          Gio.File.new_for_uri("resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita-sepia-hc.css"),
          get_media_path('/media/css/web/adwaita-sepia.css')),
]



@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/ThemeSwitcher.ui')
class ThemeSwitcher(Gtk.Box):
    __gtype_name__ = "ThemeSwitcher"

    color_scheme = "light"

    system_selector = Gtk.Template.Child()
    light_selector = Gtk.Template.Child()
    sepia_selector = Gtk.Template.Child()
    dark_selector = Gtk.Template.Child()

    settings = Settings.new()

    show_system = GObject.property(type=bool, default=True)

    @GObject.Property(type=str)
    def selected_color_scheme(self):
        """Read-write integer property."""

        return self.color_scheme

    @selected_color_scheme.setter
    def selected_color_scheme(self, color_scheme):
        self.color_scheme = color_scheme

        if color_scheme == "auto":
            self.system_selector.set_active(True)
            self.style_manager.set_color_scheme(Handy.ColorScheme.PREFER_LIGHT)
        if color_scheme == "light":
            self.light_selector.set_active(True)
            self.style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
        if color_scheme == "sepia":
            self.sepia_selector.set_active(True)
            self.style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
        if color_scheme == "dark":
            self.dark_selector.set_active(True)
            self.style_manager.set_color_scheme(Handy.ColorScheme.FORCE_DARK)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.style_manager = Handy.StyleManager.get_default()

        color_scheme = self.settings.get_string("color-scheme")
        self.color_scheme = color_scheme

        self.settings.bind(
            "color-scheme",
            self,
            "selected_color_scheme",
            Gio.SettingsBindFlags.DEFAULT)

        self.style_manager.bind_property("system-supports-color-schemes",
                                         self, "show_system",
                                         GObject.BindingFlags.SYNC_CREATE)

    @Gtk.Template.Callback()
    def _on_color_scheme_changed(self, widget, paramspec):
        if self.system_selector.get_active():
            self.selected_color_scheme = "auto"
        if self.light_selector.get_active():
            self.selected_color_scheme = "light"
        if self.sepia_selector.get_active():
            self.selected_color_scheme = "sepia"
        if self.dark_selector.get_active():
            self.selected_color_scheme = "dark"



# Copyright 2022, Manuel Genovés <manuel.genoves@gmail.com>
#                 Alexander Mikhaylenko
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE
#
# Based on code by Alexander Mikhaylenko

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import GObject, Handy


class Tweener(GObject.Object):

    def __init__(self, widget, setter, value_from, value_to, duration, offset=0, setter_args = None, callback = None, callback_arg = None):
        self.widget = widget
        self.value_from = value_from
        self.value_to = value_to
        self.duration = duration
        self.offset = offset
        if setter_args is None:
            self.setter_args = []
        else:
            self.setter_args = setter_args
        self.callback = callback
        self.callback_arg = callback_arg

        self.start_time = 0
        self.tick_cb_id = 0

        self.setter = setter

        self.current_value = 0

    def stop(self):
        if (self.tick_cb_id != 0):

            self.widget.remove_tick_callback(self.tick_cb_id)
            self.tick_cb_id = 0
            if self.callback is not None:
                self.callback(self.callback_arg)

    def start(self):
        if (not Handy.get_enable_animations(self.widget) or
                not self.widget.get_mapped() or
                self.duration < 0):
            self.setter(self.value_to, *self.setter_args)
            self.widget.queue_draw()
            return

        if (self.tick_cb_id != 0):
            self.widget.remove_tick_callback(self.tick_cb_id)
            self.tick_cb_id = 0

        self.setter(self.value_from, *self.setter_args)

        self.start_time = self.widget.get_frame_clock().get_frame_time() / 1000
        self.tick_cb_id = self.widget.add_tick_callback(self.__tick_cb)

    def __tick_cb(self, widget, frame_clock):
        frame_time = frame_clock.get_frame_time() / 1000
        t = (frame_time - self.start_time - self.offset) / self.duration

        if t >= 1:
            self.current_value = self.value_to
            self.stop()
            self.tick_cb_id = 0

            self.setter(self.current_value, *self.setter_args)

            return False

        if t < 0:
            return True

        self.current_value = self.value_from\
            + (self.value_to - self.value_from)\
            * Handy.ease_out_cubic(t)

        self.setter(self.current_value, *self.setter_args)
        self.widget.queue_draw()

        return True



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
"""
Based on latex2png.py from Stuart Rackham

AUTHOR
    Written by Stuart Rackham, <srackham@gmail.com>
    The code was inspired by Kjell Magne Fauske"s code:
    http://fauskes.net/nb/htmleqII/

    See also:
    http://www.amk.ca/python/code/mt-math
    http://code.google.com/p/latexmath2png/

COPYING
    Copyright (C) 2010 Stuart Rackham. Free use of this software is
    granted under the terms of the MIT License.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""

import os
import subprocess
import tempfile


class LatexToPNG:
    TEX_HEADER = r"""\documentclass{article}
    \usepackage{amsmath}
    \usepackage{amsthm}
    \usepackage{amssymb}
    \usepackage{bm}
    \newcommand{\mx}[1]{\mathbf{\bm{#1}}} % Matrix command
    \newcommand{\vc}[1]{\mathbf{\bm{#1}}} % Vector command
    \newcommand{\T}{\text{T}}			 % Transpose
    \pagestyle{empty}
    \begin{document}"""

    TEX_FOOTER = r"""\end{document}"""

    def __init__(self):
        self.temp_result = tempfile.NamedTemporaryFile(suffix=".png")

    def latex2png(self, tex, outfile, dpi, modified):
        """Convert LaTeX input file infile to PNG file named outfile."""
        outfile = os.path.abspath(outfile)
        outdir = os.path.dirname(outfile)
        texfile = tempfile.mktemp(suffix=".tex", dir=os.path.dirname(outfile))
        basefile = os.path.splitext(texfile)[0]
        dvifile = basefile + ".dvi"
        temps = [basefile + ext for ext in (".tex", ".dvi", ".aux", ".log")]
        skip = False

        tex = "{}\n{}\n{}\n".format(self.TEX_HEADER, tex.strip(), self.TEX_FOOTER)

        open(texfile, "w").write(tex)
        saved_pwd = os.getcwd()

        os.chdir(outdir)

        args = ["latex", "-halt-on-error", texfile]
        p = subprocess.Popen(args,
                             stderr=subprocess.STDOUT,
                             stdout=subprocess.PIPE)

        output = p.stdout
        output_lines = output.readlines()
        if os.path.isfile(dvifile):  # DVI File exists
            # Convert DVI file to PNG.
            args = ["dvipng",
                    "-D", str(dpi),
                    "-T", "tight",
                    "-x", "1000",
                    "-z", "9",
                    "-bg", "Transparent",
                    "-o", outfile,
                    dvifile]

            p = subprocess.Popen(args)
            p.communicate()

        else:
            self.clean_up(temps)
            """
            Errors in Latex output start with "! "
            Stripping exclamation marks and superflous newlines
            and telling the user what he"s done wrong.
            """
            i = []
            error = ""
            for line in output_lines:
                line = line.decode("utf-8")
                if line.startswith("!"):
                    error += line[2:]  # removing "! "
            if error.endswith("\n"):
                error = error[:-1]
            raise Exception(error)

    def generatepng(self, formula):
        try:
            self.temp_result = tempfile.NamedTemporaryFile(suffix=".png")
            formula = "$" + formula + "$"
            self.latex2png(formula, self.temp_result.name, 300, False)
            return True, self.temp_result.name

        except Exception as e:
            self.temp_result.close()
            return False, e.args[0]

    def clean_up(self, files):
        for f in files:
            if os.path.isfile(f):
                os.remove(f)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# BEGIN LICENSE
# Copyright (C) 2019, Wolf Vollprecht <w.vollprecht@gmail.com>
#               2021, Manuel Genovés <manuel.genoves@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

"""Manages all the export operations and dialogs
"""

# pylint: disable=no-member

import logging
import os
from gettext import gettext as _

from zipfile import ZipFile
import json

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gtk, Gdk, Gio, GObject, Handy

from apostrophe import helpers
from apostrophe.theme_switcher import Theme

LOGGER = logging.getLogger('apostrophe')


class Format(GObject.Object):

    def __init__(self, name, ext, to, **kwargs):
        super().__init__(**kwargs)
        self.name: str = name
        self.ext: str = ext
        self.to: str = to

    @property
    def has_pages(self):
        return self.to in {"pdf", "odt", "docx"}

    @property
    def is_html(self):
        return self.to == "html5"

    @property
    def has_syntax(self):
        return self.ext in {"html", "tex", "docx", "pdf"}

    @property
    def is_presentation(self):
        return self.to in {"beamer", "revealjs", "dzslides"}

    @property
    def requires_texlive(self):
        return self.ext in {"tex", "pdf"}


class ExportDialog:

    __gtype_name__ = "ExportDialog"

    formats = {
        "pdf":
        {
            "name": "PDF",
            "extension": "pdf",
            "to": "pdf",
            "mimetype": "application/pdf",
            "args": ["--pdf-engine=xelatex",
                     "--variable=papersize:a4"]
        },
        "html":
        {
            "name": "HTML",
            "extension": "html",
            "to": "html5",
            "mimetype": "text/html",
            "args": ["--self-contained",
                     "--css=%s" % Theme.get_current().web_css,
                     "--mathjax",
                     "--lua-filter=%s"
                     % helpers.get_media_path('/lua/relative_to_absolute.lua'),
                     "--lua-filter=%s"
                     % helpers.get_media_path('/lua/task-list.lua')]
        },
        "odt":
        {
            "name": "ODT",
            "extension": "odt",
            "to": "odt",
            "mimetype": "application/vnd.oasis.opendocument.text",
            "args": ["--variable=papersize:a4"]
        }
    }

    def __init__(self, file, _format, text):
        self.format = _format
        self.text = text

        self._show_texlive_warning = (self.format == "pdf" and
                                      not helpers.exist_executable("pdftex"))

        if (self._show_texlive_warning):
            self.dialog = Gtk.MessageDialog(
                None,
                Gtk.DialogFlags.MODAL |
                Gtk.DialogFlags.DESTROY_WITH_PARENT,
                Gtk.MessageType.ERROR,
                Gtk.ButtonsType.CLOSE
            )

            self.dialog.get_message_area().add(TexliveWarning())

        else:
            self.dialog = Gtk.FileChooserNative.new(
                          _("Export"),
                          None,
                          Gtk.FileChooserAction.SAVE,
                          _("Export to %s") %
                          self.formats[self.format]["name"],
                          _("Cancel"))

            dialog_filter = Gtk.FileFilter.new()
            dialog_filter.set_name(self.formats[self.format]["name"])
            dialog_filter.add_mime_type(self.formats[self.format]["mimetype"])
            self.dialog.add_filter(dialog_filter)
            self.dialog.set_do_overwrite_confirmation(True)

            self.dialog.set_current_name(
                file.name + '.' + self.formats[self.format]["extension"])

    def export(self):
        response = self.dialog.run()

        if not self._show_texlive_warning:
            file = self.dialog.get_file()
            fmt = self.formats[self.format]["to"]
            args = self.formats[self.format]["args"]

            if response == Gtk.ResponseType.ACCEPT:
                try:
                    export(self.text, file, fmt, args)
                except (NotADirectoryError, RuntimeError) as e:
                    helpers.show_error(
                        None,
                        _("An error happened while trying to export:\n\n"
                          "{err_msg}")
                        .format(err_msg=str(e).encode()
                                .decode("unicode-escape")))

        self.dialog.destroy()


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/Export.ui')
class AdvancedExportDialog(Handy.Window):

    __gtype_name__ = "AdvancedExportDialog"

    headerbar = Gtk.Template.Child()

    formats_list = Gtk.Template.Child()

    leaflet = Gtk.Template.Child()
    options_page = Gtk.Template.Child()
    formats_page = Gtk.Template.Child()

    # #### --option properties-- #####
    sw_standalone = Gtk.Template.Child()
    sw_toc = Gtk.Template.Child()
    sw_numbers = Gtk.Template.Child()

    cmb_page_size = Gtk.Template.Child()

    sw_self_contained = Gtk.Template.Child()

    sw_syntax_highlighting = Gtk.Template.Child()
    cmb_syntax_highlighting = Gtk.Template.Child()

    sw_incremental_bullets = Gtk.Template.Child()
    # #### ---------------------- #####

    formats = Gio.ListStore.new(Format)

    page_sizes = ['A4', 'Letter']
    syntax_styles = ['pygments',
                     'kate',
                     'monochrome',
                     'espresso',
                     'zenburn',
                     'haddock',
                     'tango']

    def __init__(self, file, text, **kwargs):
        super().__init__(**kwargs)

        self.file = file
        self.text = text

        if not self.formats:
            with open(helpers.get_media_path("/media/formats.json")) as f:
                _formats_list = json.load(f)
            for _i, format in enumerate(_formats_list):
                self.formats.append(Format(format["name"],
                                           format["ext"],
                                           format["to"]))

        self.formats_list.bind_model(self.formats, self.row_constructor, None)
        self.formats_list.select_row(self.formats_list.get_row_at_index(0))

        page_sizes_list = helpers.liststore_from_list(self.page_sizes)
        self.cmb_page_size.bind_name_model(page_sizes_list,
                                           self.get_hdy_name, None, None)

        syntax_styles_list = helpers.liststore_from_list(self.syntax_styles)
        self.cmb_syntax_highlighting.bind_name_model(syntax_styles_list,
                                                     self.get_hdy_name,
                                                     None, None)

    @GObject.Property(type=str)
    def title(self):
        name = self.formats_list.get_selected_row().item.name
        return _("Export to {}").format(name)

    @GObject.Property(type=bool, default=False)
    def show_page_size_options(self):
        return self.formats_list.get_selected_row().item.has_pages

    @GObject.Property(type=bool, default=False)
    def show_html_options(self):
        return self.formats_list.get_selected_row().item.is_html

    @GObject.Property(type=bool, default=False)
    def show_syntax_options(self):
        return self.formats_list.get_selected_row().item.has_syntax

    @GObject.Property(type=bool, default=False)
    def show_presentation_options(self):
        return self.formats_list.get_selected_row().item.is_presentation

    @GObject.Property(type=bool, default=False)
    def show_texlive_warning(self):
        is_tex = self.formats_list.get_selected_row().item.requires_texlive
        texlive_installed = helpers.exist_executable("pdftex")

        return is_tex and not texlive_installed

    @GObject.Property(type=bool, default=False)
    def show_go_back_button(self):
        folded = self.leaflet.props.folded
        on_options_page = (self.leaflet.get_visible_child() ==
                           self.options_page)

        return folded and on_options_page

    @GObject.Property(type=str, default="options")
    def options_page_name(self):
        name = "texlive_warning" if self.show_texlive_warning else "options"
        return name

    @GObject.Property(type=bool, default=False)
    def exports_multiple_files(self):
        return self.formats_list.get_selected_row().item.to == "revealjs"

    def row_constructor(self, item, _user_data):
        row = Handy.ActionRow.new()
        row.item = item
        row.set_title(item.name)

        return row

    def get_hdy_name(self, item, _user_data, _user_data_free):
        return item.dup_string()

    def get_hdy_comborow_name(self, hdy_cmbrow):
        model = hdy_cmbrow.get_model()
        index = hdy_cmbrow.get_selected_index()
        item = model.get_item(index)

        return item.dup_string()

    @Gtk.Template.Callback()
    def reveal_go_back(self, _widget, *args):
        self.notify("show_go_back_button")

    @Gtk.Template.Callback()
    def go_back(self, _widget):
        self.leaflet.set_visible_child(self.formats_page)

    @Gtk.Template.Callback()
    def on_format_selected(self, _widget, _row):
        self.leaflet.set_visible_child(self.options_page)

        self.notify("show_page_size_options")
        self.notify("show_html_options")
        self.notify("show_syntax_options")
        self.notify("show_presentation_options")
        self.notify("show_texlive_warning")
        self.notify("options_page_name")
        self.notify("title")

    @Gtk.Template.Callback()
    def on_destroy(self, _widget):
        self.destroy()

    @Gtk.Template.Callback()
    def export(self, widget):
        self.retrieve_args()

        if self.exports_multiple_files:
            export_dialog = Gtk.FileChooserNative.new(
                _("Export"), None, Gtk.FileChooserAction.SELECT_FOLDER,
                _("Select folder"), _("Cancel"))
        else:
            export_dialog = Gtk.FileChooserNative.new(
                _("Export"), None,  Gtk.FileChooserAction.SAVE,
                _("Export to %s") %
                self.formats_list.get_selected_row().item.name, _("Cancel"))

            export_dialog.set_current_name(
                self.file.name + '.' +
                self.formats_list.get_selected_row().item.ext)

        export_dialog.set_transient_for(self)
        export_dialog.set_do_overwrite_confirmation(True)

        response = export_dialog.run()
        if self.exports_multiple_files:
            folder = export_dialog.get_file()
            with ZipFile(helpers.get_media_path("/media/reveal.js.zip"),
                         "r") as zipObj:
                zipObj.extractall(folder.get_path())
            export_file = folder.get_child(self.file.name + '.' +
                              self.formats_list.get_selected_row().item.ext)
        else:
            export_file = export_dialog.get_file()

        fmt = self.formats_list.get_selected_row().item.to
        args = self.retrieve_args()

        if response == Gtk.ResponseType.ACCEPT:
            try:
                export(self.text, export_file, fmt, args)
            except (NotADirectoryError, RuntimeError) as e:
                helpers.show_error(
                    None,
                    _("An error happened while trying to export:\n\n{err_msg}")
                    .format(err_msg=str(e).encode().decode("unicode-escape")))

        export_dialog.destroy()
        self.destroy()

    def retrieve_args(self):
        args = []

        if self.formats_list.get_selected_row().item.ext == "pdf":
            args.append("--pdf-engine=xelatex")

        if self.sw_standalone.get_active():
            args.append("--standalone")
        if self.sw_toc.get_active():
            args.append("--toc")
        if self.sw_numbers.get_active():
            args.append("--number-sections")

        if (self.show_page_size_options and
           self.cmb_page_size.get_selected_index() == 0):
            if ((fmt := self.formats_list.get_selected_row().item.to) in
               {"pdf", "latex", "context"}):
                args.append("--variable=papersize:a4")
            elif fmt in ("odt", "docx"):
                args.append("--reference-doc=" + helpers.get_media_path(
                    "/reference_files/reference-a4." + fmt))

        if self.show_html_options:
            args.append("--css=%s" % Theme.get_current().web_css)
            args.append("--mathjax")
            args.append("--lua-filter=%s" % helpers.get_media_path(
                '/lua/relative_to_absolute.lua'))
            args.append("--lua-filter=%s" % helpers.get_media_path(
                '/lua/task-list.lua'))
            if self.sw_self_contained.get_active():
                args.append("--self-contained")

        if self.show_syntax_options:
            if self.sw_syntax_highlighting.get_enable_expansion():
                selected_style = self.get_hdy_comborow_name(
                    self.cmb_syntax_highlighting)
                args.append("--highlight-style={}".format(selected_style))

        if self.show_presentation_options:
            if self.sw_incremental_bullets.get_active():
                args.append("--incremental")

        if self.formats_list.get_selected_row().item.to == "revealjs":
            args.extend(["-V", "revealjs-url=reveal.js"])

        return args


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/TexliveWarning.ui')
class TexliveWarning(Gtk.Stack):

    __gtype_name__ = 'TexliveWarning'
    command = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        child_name = "flatpak" if os.path.isfile("/.flatpak-info") else "distro"
        self.set_visible_child_name(child_name)

    @Gtk.Template.Callback()
    def copy(self, _widget):
        cb = Gtk.Clipboard.get_default(Gdk.Display.get_default())
        cb.set_text(self.command.get_text(), -1)


def export(text, file, _format, args):
    """Export the given text using the specified format.
    For advanced export, this includes special flags for the enabled options.

    Keyword Arguments:
        text {str} -- Text to export (default: {""})
    """

    to = _format

    helpers.pandoc_convert(
        text, to=to, args=args, outputfile=file.get_path())



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import gi

from apostrophe.helpers import user_action
from apostrophe.inline_preview import InlinePreview
from apostrophe.text_view_drag_drop_handler import DragDropHandler,\
                                                   TARGET_URI, TARGET_TEXT
from apostrophe.text_view_format_inserter import FormatInserter
from apostrophe.text_view_markup_handler import MarkupHandler
from apostrophe.text_view_scroller import TextViewScroller
from apostrophe.text_view_undo_redo_handler import UndoRedoHandler

gi.require_version('Gtk', '3.0')
gi.require_version('Gspell', '1')
from gi.repository import Gtk, Gdk, GObject, GLib, Gspell

import logging

LOGGER = logging.getLogger('apostrophe')

DEFAULT_DPI = 98304

@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/TextView.ui')
class ApostropheTextView(Gtk.TextView):
    """ApostropheTextView encapsulates all the features around the editor.

    It combines the following:
    - Undo / redo (via TextBufferUndoRedoHandler)
    - Format shortcuts (via TextBufferShortcutInserter)
    - Markup (via TextBufferMarkupHandler)
    - Preview popover (via TextBufferMarkupHandler)
    - Drag and drop (via TextViewDragDropHandler)
    - Scrolling (via TextViewScroller)
    - The various modes supported by Apostrophe (e. Focus Mode, Hemingway Mode)
    """

    __gtype_name__ = "ApostropheTextView"

    __gsignals__ = {
        'insert-italic': (GObject.SignalFlags.ACTION, None, ()),
        'insert-bold': (GObject.SignalFlags.ACTION, None, ()),
        'insert-hrule': (GObject.SignalFlags.ACTION, None, ()),
        'insert-listitem': (GObject.SignalFlags.ACTION, None, ()),
        'insert-header': (GObject.SignalFlags.ACTION, None, ()),
        'insert-strikethrough': (GObject.SignalFlags.ACTION, None, ()),
        'undo': (GObject.SignalFlags.ACTION, None, ()),
        'redo': (GObject.SignalFlags.ACTION, None, ()),
        'scroll-scale-changed': (GObject.SIGNAL_RUN_LAST, None, (float,)),
    }


    bigger_text = GObject.Property(type=bool, default=False)
    hemingway_mode = GObject.Property(type=bool, default=False)
    focus_mode = GObject.Property(type=bool, default=False)
    font_size = GObject.Property(type=int, default=16)
    line_chars = GObject.Property(type=int, default=66)
    spellcheck = GObject.Property(type=bool, default=True)

    scroller = None

    _scroll_scale = 0

    @GObject.Property(type=float, default=0)
    def scroll_scale(self):
        return self.scroller.get_scroll_scale() if self.scroller else 0

    @scroll_scale.setter
    def scroll_scale(self, scale):
        if self.scroller:
            self.scroller.set_scroll_scale(scale)

    def __init__(self):
        super().__init__()

        # Spell checking
        self.gspell_view = Gspell.TextView.get_from_gtk_text_view(self)
        self.gspell_view.basic_setup()

        # Undo / redo
        self.undo_redo = UndoRedoHandler()
        self.get_buffer().connect('begin-user-action',
                                  self.undo_redo.on_begin_user_action)
        self.get_buffer().connect('end-user-action',
                                  self.undo_redo.on_end_user_action)
        self.get_buffer().connect('insert-text',
                                  self.undo_redo.on_insert_text)
        self.get_buffer().connect('delete-range',
                                  self.undo_redo.on_delete_range)
        self.connect('undo', self.undo_redo.undo)
        self.connect('redo', self.undo_redo.redo)

        # Format shortcuts
        self.shortcut = FormatInserter()
        self.connect('insert-italic', self.shortcut.insert_italic)
        self.connect('insert-bold', self.shortcut.insert_bold)
        self.connect('insert-strikethrough', self.shortcut.insert_strikethrough)
        self.connect('insert-hrule', self.shortcut.insert_horizontal_rule)
        self.connect('insert-listitem', self.shortcut.insert_list_item)
        self.connect('insert-header', self.shortcut.insert_header)

        # Markup
        self.markup = MarkupHandler(self)
        self.connect('style-updated', self.markup.on_style_updated)
        self.connect('destroy', self.markup.stop)

        # Preview popover
        self.preview_popover = InlinePreview(self)

        # Drag and drop
        self.drag_drop = DragDropHandler(self, TARGET_URI, TARGET_TEXT)

        # While resizing the TextView, there is unwanted scroll upwards
        # if a top margin is present.
        # When a size allocation is detected, this variable will hold
        # the scroll to re-set until the UI is idle again.

        # TODO: Find a better way to handle unwanted scroll.
        self.frozen_scroll_scale = None

    def get_text(self):
        text_buffer = self.get_buffer()
        start_iter = text_buffer.get_start_iter()
        end_iter = text_buffer.get_end_iter()
        return text_buffer.get_text(start_iter, end_iter, False)

    def set_text(self, text):
        """Set text and clear undo history"""

        text_buffer = self.get_buffer()
        with user_action(text_buffer):
            text_buffer.set_text(text)
        self.undo_redo.clear()

    def update_vertical_margin(self, top_margin=0):
        if self.focus_mode:
            height = self.get_allocation().height + top_margin 
            # The height/6 may seem strange. It's a workaround for a GTK bug
            # If a top-margin is larger than a certain amount of
            # the TextView height, jumps may occur when pressing enter.

            # TODO: check again in GTK4
            self.set_top_margin(height / 2 + top_margin - height/6)
            self.set_bottom_margin(height / 2)
        else:
            self.props.top_margin = 80 + top_margin
            self.props.bottom_margin = 64

    def clear(self):
        """Clear text and undo history"""

        self.set_text('')

    def smooth_scroll_to(self, mark=None):
        """Scrolls if needed to ensure mark is visible.

        If mark is unspecified, the cursor is used."""

        if self.scroller is None:
            return
        if mark is None:
            mark = self.get_buffer().get_insert()
        GLib.idle_add(self.scroller.smooth_scroll_to_mark,
                      mark, self.focus_mode)

    @Gtk.Template.Callback()
    def _on_button_release_event(self, _widget, _event):
        if self.focus_mode:
            self.markup.apply()
        return False

    @Gtk.Template.Callback()
    def _on_focus_mode_update(self, *args, **kwargs):
        self.update_vertical_margin()
        self.markup.apply()
        self.smooth_scroll_to()
        self._on_spellcheck_update()
        self.grab_focus()

    @Gtk.Template.Callback()
    def _on_key_press_event(self, _widget, event):
        if self.hemingway_mode:
            return event.keyval == Gdk.KEY_BackSpace or event.keyval == Gdk.KEY_Delete

        if event.state & Gdk.ModifierType.SHIFT_MASK == Gdk.ModifierType.SHIFT_MASK \
                and event.keyval == Gdk.KEY_ISO_Left_Tab:  # Capure Shift-Tab
            self._on_shift_tab()
            return True

    @Gtk.Template.Callback()
    def _on_mark_set(self, _text_buffer, _location, mark, _data=None):
        if mark.get_name() == 'selection_bound':
            self.markup.apply()
            if not self.get_buffer().get_has_selection():
                self.smooth_scroll_to(mark)
        elif mark.get_name() == 'gtk_drag_target':
            self.smooth_scroll_to(mark)
        return True

    # TODO: this seems a little backwards way to hook things up
    @Gtk.Template.Callback()
    def _on_parent_set(self, *_):
        parent = self.get_parent()
        if parent:
            parent.set_size_request(self.get_min_width(), 500)
            self.scroller = TextViewScroller(self, parent)
            parent.get_vadjustment().connect("changed",
                                             self._on_vadjustment_changed)
            parent.get_vadjustment().connect("value-changed",
                                             self._on_vadjustment_changed)
        else:
            self.scroller = None

    @Gtk.Template.Callback()
    def _on_paste_done(self, *_):
        self.smooth_scroll_to()

    def _on_shift_tab(self):
        """Delete last character if it is a tab"""
        text_buffer = self.get_buffer()
        pen_iter = text_buffer.get_end_iter()
        pen_iter.backward_char()
        end_iter = text_buffer.get_end_iter()

        if pen_iter.get_char() == "\t":
            with user_action(text_buffer):
                text_buffer.delete(pen_iter, end_iter)

    @Gtk.Template.Callback()
    def _on_size_allocate(self, *_):
        self._update_horizontal_margin()
        self.markup.update_margins_indents()
        self.queue_draw()

        # TODO: Find a better way to handle unwanted scroll on resize.
        self.frozen_scroll_scale = self.scroll_scale
        GLib.idle_add(self._unfreeze_scroll_scale)

    @Gtk.Template.Callback()
    def _on_spellcheck_update(self, *args, **kwargs):
        self.gspell_view.set_inline_spell_checking(
            self.spellcheck and not self.focus_mode)

    @Gtk.Template.Callback()
    def _on_text_changed(self, *_):
        self.markup.apply()
        self.smooth_scroll_to()

    def _on_vadjustment_changed(self, *_):
        if self.frozen_scroll_scale is not None:
            self.scroll_scale = self.frozen_scroll_scale
        elif self.scroller.can_scroll():
            self.emit("scroll-scale-changed", self.scroll_scale)

    def _unfreeze_scroll_scale(self):
        self.frozen_scroll_scale = None
        self.queue_draw()

    @Gtk.Template.Callback()
    def _update_horizontal_margin(self, *args, **kwargs):
        width = self.get_allocation().width
        # Ensure the appropriate font size is being used
        for font_size in self._get_font_sizes():
            if width >= self.get_min_width(font_size):
                if font_size != self.font_size:
                    self.font_size = font_size
                    style_ctxt = self.get_style_context()
                    for size_class in filter(lambda style_class: style_class.startswith("size"), style_ctxt.list_classes()):
                        style_ctxt.remove_class(size_class)
                    style_ctxt.add_class("size{}".format(font_size))
                break
        else:
            return

        # Apply margin with the remaining space to allow for markup
        line_width = (self.line_chars + 1) *\
            int(self._get_char_width(self.font_size)) - 1
        horizontal_margin = (width - line_width) / 2
        self.props.left_margin = horizontal_margin
        self.props.right_margin = horizontal_margin

    def _get_font_sizes(self):
        font_sizes_list = [20, 18, 17, 16, 15, 14]
        if self.bigger_text:
            font_sizes_list[:0] = [24, 22]

        return font_sizes_list

    def get_min_width(self, font_size=None):
        """Returns the minimum width of this text view."""

        if font_size is None:
            font_size = self._get_font_sizes()[-1]
        return (self.line_chars + self._get_pad_chars(font_size) + 1) \
            * self._get_char_width(font_size) - 1

    def _get_pad_chars(self, font_size):
        """Returns the amount of character padding for font_size.

        Markup can use up to 7 in normal conditions."""

        if self.bigger_text:
            return 8 * int((1 + (font_size - self._get_font_sizes()[-1])/3))
        else:
            return 8 * (1 + font_size - self._get_font_sizes()[-1])

    @staticmethod
    def _get_char_width(font_size):
        """Returns the font width for a given size.
        Note: specific to Fira Mono!"""

        # We scale by the text scale factor system-wide setting.
        # In the css the units are in em so they're automatically scaled
        # 14.666px == 1em

        settings = Gtk.Settings.get_default()
        scale_factor = settings.props.gtk_xft_dpi / DEFAULT_DPI

        return scale_factor * font_size * 1 / 1.6



from queue import Queue
from threading import Thread

from gi.repository import GLib

from apostrophe import helpers
from apostrophe.theme_switcher import Theme, ThemeSwitcher


class PreviewConverter:
    """Converts markdown to html using a background thread."""

    def __init__(self):
        super().__init__()

        self.queue = Queue()
        worker = Thread(target=self.__do_convert, name="preview-converter")
        worker.daemon = True
        worker.start()

    def convert(self, text, callback, *user_data):
        """Converts text to html, calling callback when done.

        The callback argument contains the result."""

        self.queue.put((text, callback, user_data))

    def stop(self):
        """Stops the background worker.
        PreviewConverter shouldn't be used after this."""

        self.queue.put((None, None))

    def __do_convert(self):
        while True:
            while True:
                (text, callback, user_data) = self.queue.get()
                if text is None and callback is None:
                    return
                if self.queue.empty():
                    break

            args = ['--standalone',
                    '--mathjax',
                    '--css=' + Theme.get_current().web_css,
                    '--lua-filter=' +
                    helpers.get_media_path('/lua/relative_to_absolute.lua'),
                    '--lua-filter=' +
                    helpers.get_media_path('/lua/task-list.lua')]
            text = helpers.pandoc_convert(text, to="html5", args=args)

            GLib.idle_add(callback, text, *user_data)



import mimetypes
import urllib
from gettext import gettext as _
from os.path import basename
from apostrophe.settings import Settings

from gi.repository import Gtk

(TARGET_URI, TARGET_TEXT) = range(2)


class DragDropHandler:
    TARGET_URI = None

    def __init__(self, text_view, *targets):
        super().__init__()

        self.settings = Settings.new()

        self.target_list = Gtk.TargetList.new([])
        if TARGET_URI in targets:
            self.target_list.add_uri_targets(TARGET_URI)
        if TARGET_TEXT in targets:
            self.target_list.add_text_targets(TARGET_TEXT)

        text_view.drag_dest_set_target_list(self.target_list)
        text_view.connect_after(
            'drag-data-received',
            self.on_drag_data_received)

    def on_drag_data_received(
            self, text_view, drag_context, _x, _y, data, info, time):
        """Handle drag and drop events"""

        text_buffer = text_view.get_buffer()

        if info == TARGET_URI:
            uris = data.get_uris()
            for uri in uris:
                name = basename(urllib.parse.unquote_plus(uri))
                mime = mimetypes.guess_type(uri)

                if mime[0] is not None and mime[0].startswith('image/'):
                    basepath = self.settings.get_string("open-file-path")
                    basepath = urllib.parse.quote(basepath)

                    if uri.startswith("file://"):
                        uri = uri[7:]

                    # for handling local URIs we need to substract the basepath
                    # except when it is "/" (document not saved)
                    if uri.startswith(basepath) and basepath != "/":
                        uri = uri[len(basepath) + 1:]

                    text = "![{}]({})".format(name, uri)
                    limit_left = 2
                    limit_right = len(name)
                else:
                    text = "[{}]({})".format(name, uri)
                    limit_left = 1
                    limit_right = len(name)

        elif info == TARGET_TEXT:
            text = data.get_text()

            # delete automatically added DnD text
            insert_mark = text_buffer.get_insert()
            cursor_iter_r = text_buffer.get_iter_at_mark(insert_mark)
            cursor_iter_l = cursor_iter_r.copy()
            cursor_iter_l.backward_chars(len(data.get_text()))

            text_buffer.delete(cursor_iter_l, cursor_iter_r)

            if text.startswith(("http://", "https://", "www.")):
                text = "[{}]({})".format(_("web page"), text)
                limit_left = 1
                limit_right = len(_("web page"))
            else:
                limit_left = 0
                limit_right = 0

        text_buffer.place_cursor(text_buffer.get_iter_at_mark(
            text_buffer.get_mark('gtk_drag_target')))
        text_buffer.insert_at_cursor(text)
        insert_mark = text_buffer.get_insert()
        selection_bound = text_buffer.get_selection_bound()
        cursor_iter = text_buffer.get_iter_at_mark(insert_mark)
        cursor_iter.backward_chars(len(text) - limit_left)
        text_buffer.move_mark(insert_mark, cursor_iter)
        cursor_iter.forward_chars(limit_right)
        text_buffer.move_mark(selection_bound, cursor_iter)

        Gtk.drag_finish(drag_context, True, True, time)
        text_view.get_toplevel().present_with_time(time)
        return False



import logging

LOGGER = logging.getLogger('apostrophe')


class UndoableInsert:
    """Something has been inserted into text_buffer"""

    def __init__(self, text_iter, text, length):
        self.offset = text_iter.get_offset()
        self.text = text
        self.length = length
        self.mergeable = not bool(
            self.length > 1 or self.text in (
                "\r", "\n", " "))

    def undo(self, text_buffer):
        offset = self.offset
        start = text_buffer.get_iter_at_offset(offset)
        stop = text_buffer.get_iter_at_offset(offset + self.length)
        text_buffer.place_cursor(start)
        text_buffer.delete(start, stop)

    def redo(self, text_buffer):
        start = text_buffer.get_iter_at_offset(self.offset)
        text_buffer.insert(start, self.text)
        new_cursor_pos = text_buffer.get_iter_at_offset(
            self.offset + self.length)
        text_buffer.place_cursor(new_cursor_pos)

    def merge(self, next_action):
        """Merge a following action into this insert, if possible

        can't merge if prev is not another insert
        can't merge if prev and cur are not mergeable in the first place
        can't merge when user set the input bar somewhere else
        can't merge across word boundaries"""

        if not isinstance(next_action, UndoableInsert):
            return False
        if not self.mergeable or not next_action.mergeable:
            return False
        if self.offset + self.length != next_action.offset:
            return False
        whitespace = (' ', '\t')
        if self.text in whitespace != next_action.text in whitespace:
            return False

        self.length += next_action.length
        self.text += next_action.text
        return True


class UndoableDelete:
    """Something has been deleted from text_buffer"""

    def __init__(self, text_buffer, start_iter, end_iter):
        self.text = text_buffer.get_text(start_iter, end_iter, False)
        self.start = start_iter.get_offset()
        self.end = end_iter.get_offset()
        # Find out if backspace or delete were used to not mess up redo
        insert_iter = text_buffer.get_iter_at_mark(text_buffer.get_insert())
        self.delete_key_used = bool(insert_iter.get_offset() <= self.start)
        self.mergeable = not bool(
            self.end -
            self.start > 1 or self.text in (
                "\r",
                "\n",
                " "))

    def undo(self, text_buffer):
        start = text_buffer.get_iter_at_offset(self.start)
        text_buffer.insert(start, self.text)
        if self.delete_key_used:
            text_buffer.place_cursor(start)
        else:
            stop = text_buffer.get_iter_at_offset(self.end)
            text_buffer.place_cursor(stop)

    def redo(self, text_buffer):
        start = text_buffer.get_iter_at_offset(self.start)
        stop = text_buffer.get_iter_at_offset(self.end)
        text_buffer.delete(start, stop)
        text_buffer.place_cursor(start)

    def merge(self, next_action):
        """Check if this delete can be merged with a previous action

        can't merge if prev is not another delete
        can't merge if prev and cur are not mergeable in the first place
        can't merge if delete and backspace key were both used
        can't merge across word boundaries"""

        if not isinstance(next_action, UndoableDelete):
            return False
        if not self.mergeable or not next_action.mergeable:
            return False
        if self.delete_key_used != next_action.delete_key_used:
            return False
        if self.start != next_action.start and self.start != next_action.end:
            return False
        whitespace = (' ', '\t')
        if self.text in whitespace != next_action.text in whitespace:
            return False

        if self.delete_key_used:
            self.text += next_action.text
            self.end += (next_action.end - next_action.start)
        else:
            self.text = "%s%s" % (next_action.text, self.text)
            self.start = next_action.start
        return True


class UndoableGroup(list):
    """A list of undoable actions, usually corresponding to a single user action"""

    def undo(self, text_buffer):
        for undoable in reversed(self):
            undoable.undo(text_buffer)

    def redo(self, text_buffer):
        for undoable in self:
            undoable.redo(text_buffer)

    def merge(self, next_action):
        if len(self) == 1:
            return self[0].merge(next_action)
        else:
            return False


class UndoRedoHandler:
    """Manages undo/redo for a given text_buffer.

    Methods can be called directly, as well as be used as signal callbacks."""

    def __init__(self):
        self.undo_stack = []
        self.redo_stack = []
        self.current_undo_group = None
        self.undo_in_progress = False

    def undo(self, text_view, _data=None):
        """Undo insertions or deletions. Undone actions are moved to redo stack.

        This method can be registered to a custom undo signal,
        or used independently."""

        if not self.undo_stack:
            return
        self.undo_in_progress = True
        undo_action = self.undo_stack.pop()
        self.redo_stack.append(undo_action)
        undo_action.undo(text_view.get_buffer())
        self.undo_in_progress = False

    def redo(self, text_view, _data=None):
        """Redo insertions or deletions. Redone actions are moved to undo stack

        This method can be registered to a custom redo signal,
        or used independently."""

        if not self.redo_stack:
            return
        self.undo_in_progress = True
        redo_action = self.redo_stack.pop()
        self.undo_stack.append(redo_action)
        redo_action.redo(text_view.get_buffer())
        self.undo_in_progress = False

    def clear(self):
        self.undo_stack = []
        self.redo_stack = []

    def on_begin_user_action(self, _text_buffer):
        """Start of a user action.
        Refer to TextBuffer's "begin-user-action" signal.

        This method must be registered to TextBuffer's
        "begin-user-action" signal, or called
        manually followed by on_end_user_action."""

        self.current_undo_group = UndoableGroup()

    def on_end_user_action(self, _text_buffer):
        """End of a user action.
        Refer to TextBuffer's "end-user-action" signal.

        This method must be registered to TextBuffer's
        "end-user-action" signal, or called
        manually preceded by on_start_user_action."""

        if self.current_undo_group:
            self.undo_stack.append(self.current_undo_group)
        self.current_undo_group = None

    def on_insert_text(self, _text_buffer, text_iter, text, _length):
        """Records a text insert.
        Refer to TextBuffer's "insert-text" signal.

        This method must be registered to TextBuffer's
         "insert-text" signal, or called manually
        in between on_begin_user_action and on_end_user_action."""

        self.__record_undoable(UndoableInsert(text_iter, text, len(text)))

    def on_delete_range(self, text_buffer, start_iter, end_iter):
        """Records a range deletion.
        Refer to TextBuffer's "delete-range" signal.

        This method must be registered to TextBuffer's
         "delete-range" signal, or called manually
        in between on_begin_user_action and on_end_user_action."""

        self.__record_undoable(
            UndoableDelete(
                text_buffer,
                start_iter,
                end_iter))

    def __record_undoable(self, undoable):
        """Records a change, merging it to a previous one if possible."""

        if not self.undo_in_progress:
            self.redo_stack = []
        else:
            return

        prev_group_undoable = self.current_undo_group[-1] if self.current_undo_group else None
        prev_stack_undoable = self.undo_stack[-1] if self.undo_stack else None

        if prev_group_undoable:
            merged = prev_group_undoable.merge(undoable)
        elif prev_stack_undoable:
            merged = prev_stack_undoable.merge(undoable)
        else:
            merged = False

        if not merged:
            if self.current_undo_group is None:
                LOGGER.warning("Recording a change without a user action.")
                self.undo_stack.append(undoable)
            else:
                self.current_undo_group.append(undoable)



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE
"""Manage all the headerbars related stuff
"""

import gi

from gettext import gettext as _

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Handy
from .settings import Settings
from .theme_switcher import ThemeSwitcher
from .preview_layout_switcher import PreviewLayoutSwitcher
from .open_popover import ApostropheOpenPopover


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/Headerbar.ui')
class BaseHeaderbar(Gtk.Revealer):

    __gtype_name__ = "BaseHeaderbar"

    """Base class for all headerbars
    """

    headerbar = Gtk.Template.Child()
    tutorial_button = Gtk.Template.Child()
    preview_layout_switcher = Gtk.Template.Child()

    is_fullscreen = GObject.property(type=bool, default=False)
    title = GObject.Property(type=str)
    subtitle = GObject.Property(type=str)

    @GObject.Property(type=int)
    def allocated_height(self):
        return self.get_allocated_height()

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.bind_property("is-fullscreen", self.headerbar, "show-close-button", 6)

        tutorial = GLib.Variant.new_string("resource:///org/gnome/gitlab/somas/"
                                           "Apostrophe/media/apostrophe_markdown.md")
        self.tutorial_button.set_action_target_value(tutorial)

        self.settings = Settings.new()

        #self.select_preview_layout_row()
    @Gtk.Template.Callback()
    def on_show_hide(self, widget, event):
        ''' The crossfade animation doesn't hide the alocated space
            for the headerbar. This prevents having an empty space there
        '''
        if not self.get_child_revealed():
            self.set_visible(False)



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import webbrowser

import gi

gi.require_version('WebKit2', '4.0')
from gi.repository import WebKit2, GLib, GObject


class PreviewWebView(WebKit2.WebView):
    """A WebView that provides read/write access to scroll.

    It does so using JavaScript, by continuously monitoring it while loaded.
    The alternative is using a WebExtension and C-bindings (see reference),
    but that is more complicated implementation-wise,
    as well as build-wise until we start building with Meson.

    Reference: https://github.com/aperezdc/webkit2gtk-python-webextension-example
    """

    SYNC_SCROLL_SCALE_JS = """
scale = {:.16f};
write = {};

// Configure MathJax.
if (typeof hasMathJax === "undefined") {{

    hasMathJax = false;
    if (typeof MathJax !== "undefined") {{
        hasMathJax = typeof MathJax.Hub !== "undefined";
    }}
    
    if (hasMathJax) {{
        MathJax.Hub.Config({{ messageStyle: "none" }});
    }}
}}

// Figure out if scrollable and rendered.
e = document.documentElement;
canScroll = e.scrollHeight > e.clientHeight;
wasRendered = typeof isRendered !== "undefined" && isRendered;
isRendered = wasRendered ||
        !hasMathJax ||
        MathJax.Hub.queue.running == 0 && MathJax.Hub.queue.pending == 0;

// Write the current scroll if instructed or if it was just rendered.
if (canScroll && (write || isRendered && !wasRendered)) {{
    e.scrollTop = (e.scrollHeight - e.clientHeight) * scale;
}}

// Return the current scroll if scrollable and rendered, or -1.
if (canScroll && isRendered) {{
    e.scrollTop / (e.scrollHeight - e.clientHeight);
}} else {{
    -1;
}}
""".strip()

    __gsignals__ = {
        "scroll-scale-changed": (GObject.SIGNAL_RUN_LAST, None, (float,)),
    }

    def __init__(self):
        super().__init__()

        self.connect("size-allocate", self.on_size_allocate)
        self.connect("decide-policy", self.on_decide_policy)
        self.connect("load-changed", self.on_load_changed)
        self.connect("load-failed", self.on_load_failed)
        self.connect("destroy", self.on_destroy)

        self.props.expand = True

        self.scroll_scale = -1

        self.state_loaded = False
        self.state_load_failed = False
        self.state_discard_read = False
        self.state_dirty = False
        self.state_waiting = False

        self.timeout_id = None

    def can_scroll(self):
        return self.scroll_scale != -1

    def get_scroll_scale(self):
        return self.scroll_scale

    def set_scroll_scale(self, scale):
        self.state_dirty = scale != self.scroll_scale
        self.scroll_scale = scale
        self.state_loop()

    def on_size_allocate(self, *_):
        self.set_scroll_scale(self.scroll_scale)

    def on_decide_policy(self, _web_view, decision, decision_type):
        if decision_type == WebKit2.PolicyDecisionType.NAVIGATION_ACTION and \
                decision.get_navigation_action().is_user_gesture():
            webbrowser.open(decision.get_request().get_uri())
            decision.ignore()       # Do not follow the link in the WebView
            return True
        return False

    def on_load_changed(self, _web_view, event):
        self.state_loaded = event >= WebKit2.LoadEvent.COMMITTED and not self.state_load_failed
        self.state_load_failed = False
        self.state_discard_read = event == WebKit2.LoadEvent.STARTED and self.state_waiting
        self.state_dirty = True
        self.state_loop()

    def on_load_failed(self, _web_view, _event):
        self.state_loaded = False
        self.state_load_failed = True
        self.state_loop()

    def on_destroy(self, _widget):
        self.state_loaded = False
        self.state_loop()

    def sync_scroll_scale(self, scroll_scale, write):
        self.state_waiting = True
        self.run_javascript(
            self.SYNC_SCROLL_SCALE_JS.format(
                scroll_scale, "true" if write else "false"),
            None, self.finish_sync_scroll_scale)

    def finish_sync_scroll_scale(self, _web_view, result):
        self.state_waiting = False
        result = self.run_javascript_finish(result)
        self.state_loop(result.get_js_value().to_double())

    def state_loop(self, scroll_scale=None, delay=16):  # 16ms ~ 60hz
        # Remove any pending callbacks
        if self.timeout_id:
            GLib.source_remove(self.timeout_id)
            self.timeout_id = None

        # Set scroll scale if specified, and the state is not dirty
        if not self.state_discard_read and scroll_scale not in (
                None, self.scroll_scale):
            self.scroll_scale = scroll_scale
            if self.scroll_scale != -1:
                self.emit("scroll-scale-changed", self.scroll_scale)
        self.state_discard_read = False

        # Handle the current state
        if not self.state_loaded or self.state_load_failed or self.state_waiting:
            return
        elif self.state_dirty or delay == 0:
            self.sync_scroll_scale(self.scroll_scale, self.state_dirty)
            self.state_dirty = False
        else:
            self.timeout_id = GLib.timeout_add(delay, self.state_loop, None, 0)



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

"""Helpers for the application."""
import logging
import os
import shutil
from contextlib import contextmanager

from typing import List

from gettext import gettext as _

import gi
import pypandoc
from gi.overrides.Pango import Pango

gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gtk, Gio, Handy  # pylint: disable=E0611

from apostrophe.settings import Settings
from apostrophe import config

__apostrophe_data_directory__ = '../data/'

App = Gio.Application.get_default


@contextmanager
def user_action(text_buffer):
    text_buffer.begin_user_action()
    yield text_buffer
    text_buffer.end_user_action()


def get_media_path(path):
    """Return the full path of a given path under the media dir
       (doesn't start with file:///)
    """
    media_path = "{}{}".format(config.PKGDATA_DIR, path)
    if not os.path.exists(media_path):
        media_path = None
    return media_path


class NullHandler(logging.Handler):
    def emit(self, _record):
        pass


def set_up_logging(level):
    # add a handler to prevent basicConfig
    root = logging.getLogger()
    null_handler = NullHandler()
    root.addHandler(null_handler)

    formatter = logging.Formatter(
        "%(levelname)s:%(name)s: %(funcName)s() '%(message)s'")

    logger = logging.getLogger('apostrophe')
    logger_sh = logging.StreamHandler()
    logger_sh.setFormatter(formatter)
    logger.addHandler(logger_sh)

    # Set the logging level to show debug messages.
    if level == 1:
        logger.setLevel(logging.DEBUG)
        logger.debug('logging enabled')


def show_error(parent, message):
    dialog = Gtk.MessageDialog(
        parent, Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, message)

    dialog.set_title(_("Error"))
    dialog.run()
    dialog.destroy()


def exist_executable(command):
    """return if a command can be executed in the SO

    Arguments:
        command {str} -- a command

    Returns:
        {bool} -- if the given command exists in the system
    """

    return shutil.which(command) is not None


def liststore_from_list(str_list: List[str]):
    """return a Gtk.ListStore object of Handy.TypeValues
       constructed after a list of strings

        Arguments:
            str_list {List[str]} -- a list of strings

        Returns:
            {Gtk.ListStore} -- a ListStore of Handy.ValueObjects
    """

    list_store = Gio.ListStore.new(Handy.ValueObject)

    for element in str_list:
        obj = Handy.ValueObject.new(element)
        list_store.append(obj)

    return list_store


def get_char_width(widget):
    return Pango.units_to_double(
        widget.get_pango_context().get_metrics().get_approximate_char_width())


def pandoc_convert(text, to="html5", args=[], outputfile=None):
    fr = Settings.new().get_value('input-format').get_string() or "markdown"
    # args.extend(["--quiet"])
    return pypandoc.convert_text(
        text, to, fr, extra_args=args, outputfile=outputfile)



# Copyright (C) 2022 Manuel Genovés <manuel.genoves@gmail.com>
# Copyright (C) 2017 Jason Gray <jasonlevigray3@gmail.com>
# Copyright (C) 2017 Franz Dietrich <dietrich@teilgedanken.de>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from gettext import gettext as _

from apostrophe.helpers import App


class Inhibitor:

    def __init__(self):
        self.__cookie = 0

    def inhibit(self, flags):
        """
            Disable flags
            @param flags as Gtk.ApplicationInhibitFlags
        """
        if not self.__cookie:
            self.__cookie = App().inhibit(
                App().get_active_window(),
                flags,
                _("Unsaved documents"))

    def uninhibit(self):
        """
            Remove all the powermanagement settings
        """
        if self.__cookie:
            App().uninhibit(self.__cookie)
            self.__cookie = 0



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import logging
import re

import gi

from apostrophe.helpers import user_action

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject


LOGGER = logging.getLogger('apostrophe')

@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/SearchBar.ui')
class ApostropheSearchBar(Gtk.SearchBar):
    """
    Adds (regex) search and replace functionality to
    apostrophe
    """
    __gtype_name__ = "ApostropheSearchBar"

    replace_mode_enabled = GObject.property(type=bool, default=False)
    search_entry = Gtk.Template.Child()
    regex = Gtk.Template.Child()
    case_sensitive = Gtk.Template.Child()
    replace_entry = Gtk.Template.Child()

    def __init__(self):
        
        self.textbuffer = None

        self.matches = []
        self.active = 0

    def attach(self, textview):
        self.textview = textview
        self.textbuffer = textview.get_buffer()
        self.highlight = self.textbuffer.create_tag('search_highlight',
                                                    background="yellow")

    @Gtk.Template.Callback()
    def search_enabled(self, _widget, _data):
        if self.get_search_mode():
            self.textbuffer = self.textview.get_buffer()
            if self.textbuffer.get_has_selection():
                self.search_entry.set_text(self.textbuffer.get_slice(*self.textbuffer.get_selection_bounds(), False))
            self.search_entry.grab_focus()
            self.search()
        else:
            self.textbuffer.remove_tag(self.highlight,
                                   self.textbuffer.get_start_iter(),
                                   self.textbuffer.get_end_iter())
            self.matches = []
            self.textview.grab_focus()

    @Gtk.Template.Callback()
    def replace_enabled(self, _widget, _data):
        if not self.get_search_mode():
            self.set_search_mode(True)

    @Gtk.Template.Callback()
    def search(self, _widget=None, _data=None, scroll=True):
        if not self.textbuffer:
            return
        searchtext = self.search_entry.get_text()
        context_start = self.textbuffer.get_start_iter()
        context_end = self.textbuffer.get_end_iter()
        text = self.textbuffer.get_slice(context_start, context_end, False)

        self.textbuffer.remove_tag(self.highlight, context_start, context_end)

        # case sensitive?
        flags = False
        if not self.case_sensitive.get_active():
            flags = flags | re.I

        ## regex?
        if not self.regex.get_active():
            searchtext = re.escape(searchtext)

        matches = re.finditer(searchtext, text, flags)

        self.matches = []
        self.active = 0
        for match in matches:
            self.matches.append((match.start(), match.end()))
            start_iter = self.textbuffer.get_iter_at_offset(match.start())
            end_iter = self.textbuffer.get_iter_at_offset(match.end())
            self.textbuffer.apply_tag(self.highlight, start_iter, end_iter)
        if scroll:
            self.scrollto(self.active)
        LOGGER.debug(searchtext)

    @Gtk.Template.Callback()
    def scrolltonext(self, _widget, _data=None):
        self.scrollto(self.active + 1)

    @Gtk.Template.Callback()
    def scrolltoprev(self, _widget, _data=None):
        self.scrollto(self.active - 1)

    def scrollto(self, index):
        if not self.matches:
            return
        self.active = index % len(self.matches)

        match = self.matches[self.active]

        start_iter = self.textbuffer.get_iter_at_offset(match[0])
        end_iter = self.textbuffer.get_iter_at_offset(match[1])

        # create a mark at the start of the coincidence to scroll to it
        mark = self.textbuffer.create_mark(None, start_iter, False)
        self.textview.scroller.scroll_to_mark(mark, center=True)

        # select coincidence
        self.textbuffer.select_range(start_iter, end_iter)

    @Gtk.Template.Callback()
    def hide(self, *arg, **kwargs):
        self.set_search_mode(False)

    @Gtk.Template.Callback()
    def replace_clicked(self, _widget, _data=None):
        self.replace(self.active)

    @Gtk.Template.Callback()
    def replace_all(self, _widget=None, _data=None):
        with user_action(self.textbuffer):
            for match in reversed(self.matches):
                self.__do_replace(match)
        self.search(scroll=False)

    def replace(self, searchindex, _inloop=False):
        with user_action(self.textbuffer):
            self.__do_replace(self.matches[searchindex])
        active = self.active
        self.search(scroll=False)
        self.active = active
        self.scrollto(self.active)

    def __do_replace(self, match):
        start_iter = self.textbuffer.get_iter_at_offset(match[0])
        end_iter = self.textbuffer.get_iter_at_offset(match[1])
        self.textbuffer.delete(start_iter, end_iter)
        start_iter = self.textbuffer.get_iter_at_offset(match[0])
        self.textbuffer.insert(start_iter, self.replace_entry.get_text())

    # Since the searchbar is overlayed to the textview we need to 
    # update its margin when the searchbar appears
    @Gtk.Template.Callback()
    def update_textview_margin(self, _widget, allocation):
        self.textview.update_vertical_margin(allocation.height)


# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see &lt;http://www.gnu.org/licenses/&gt;.

import gi


gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import GLib, Gio, Gtk, Gdk, Handy

from apostrophe.main_window import MainWindow
from apostrophe.settings import Settings
from apostrophe.helpers import set_up_logging
from apostrophe.preferences_dialog import ApostrophePreferencesDialog
from apostrophe.inhibitor import Inhibitor
from apostrophe.theme_switcher import Theme


class Application(Gtk.Application):

    def __init__(self, application_id, *args, **kwargs):
        super().__init__(*args, application_id=application_id,
                         flags=Gio.ApplicationFlags.HANDLES_OPEN | Gio.ApplicationFlags.NON_UNIQUE,
                         **kwargs)

        self.add_main_option("verbose", b"v", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Verbose output", None)

        # Hardcode Adwaita to prevent issues with third party themes
        gtk_settings = Gtk.Settings.get_default()
        self._set_theme(gtk_settings)
        gtk_settings.connect("notify::gtk-theme-name", self._set_theme)
        gtk_settings.connect("notify::gtk-icon-theme-name", self._set_theme)

        # Set css theme
        css_provider_file = Gio.File.new_for_uri(
            "resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/Adwaita.css")
        self.style_provider = Gtk.CssProvider()
        self.style_provider.load_from_file(css_provider_file)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), self.style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_USER)

        # Style manager
        style_manager = Handy.StyleManager.get_default()
        style_manager.connect("notify::dark", self._set_color_scheme)
        style_manager.connect("notify::high-contrast", self._set_color_scheme)
        
        # Set editor keybindings
        # SCSS is not fit for this, so we do it in an external css file
        css_bindings_file = Gio.File.new_for_uri(
            "resource:///org/gnome/gitlab/somas/Apostrophe/media/css/gtk/bindings.css")
        self.bindings_provider = Gtk.CssProvider()
        self.bindings_provider.load_from_file(css_bindings_file)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), self.bindings_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_USER)

        # Set icons
        Gtk.IconTheme.get_default().add_resource_path(
            "/org/gnome/gitlab/somas/Apostrophe/icons"
        )

        Handy.init()

        self.windows = []
        self.settings = Settings.new()
        self.inhibitor = None
        self._application_id = application_id

    def do_startup(self, *args, **kwargs):

        Gtk.Application.do_startup(self)


        color_scheme = self.settings.get_string("color-scheme")
        action = Gio.SimpleAction.new_stateful(
                 "color_scheme", GLib.VariantType.new("s"),
                 GLib.Variant.new_string(color_scheme))
        action.connect("activate", self._set_color_scheme)
        self.add_action(action)

        action = Gio.SimpleAction.new("new_window", None)
        action.connect("activate", self.on_new_window)
        self.add_action(action)

        action = Gio.SimpleAction.new("preferences", None)
        action.connect("activate", self.on_preferences)
        self.add_action(action)

        action = Gio.SimpleAction.new("shortcuts", None)
        action.connect("activate", self.on_shortcuts)
        self.add_action(action)

        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", self.on_about)
        self.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", self.on_quit)
        self.add_action(action)

        # Stats Menu

        stat_default = self.settings.get_string("stat-default")
        action = Gio.SimpleAction.new_stateful(
                 "stat_default", GLib.VariantType.new("s"),
                 GLib.Variant.new_string(stat_default))
        action.connect("activate", self.on_stat_default)
        self.add_action(action)
        
        # Shortcuts

        # TODO: be aware that a couple of shortcuts are defined in base.css

        self.set_accels_for_action("win.focus_mode", ["<Ctl>d"])
        self.set_accels_for_action("win.hemingway_mode", ["<Ctl>t"])
        self.set_accels_for_action("win.fullscreen", ["F11"])
        self.set_accels_for_action("win.find", ["<Ctl>f"])
        self.set_accels_for_action("win.find_replace", ["<Ctl>h"])
        self.set_accels_for_action("app.spellcheck", ["F7"])

        self.set_accels_for_action("win.new", ["<Ctl>n"])
        self.set_accels_for_action("win.open", ["<Ctl>o"])
        self.set_accels_for_action("win.save", ["<Ctl>s"])
        self.set_accels_for_action("win.save_as", ["<Ctl><shift>s"])
        self.set_accels_for_action("app.quit", ["<Ctl>w", "<Ctl>q"])

        # Inhibitor
        self.inhibitor = Inhibitor()

    def do_activate(self, *args, **kwargs):

        if not self.windows:
            self.settings.connect("changed", self.on_settings_changed)

        self.windows.append(MainWindow(self))

        if self._application_id == 'org.gnome.gitlab.somas.Apostrophe.Devel':
            for window in self.windows:
                window.get_style_context().add_class('devel')

        self._set_color_scheme()
        self.windows[-1].present()

    def do_handle_local_options(self, options):
        if options.contains("verbose") or self._application_id \
                == 'org.gnome.gitlab.somas.Apostrophe.Devel':

            set_up_logging(1)
        return -1

    def do_open(self, files, _n_files, _hint):
        self.activate()
        for i, file in enumerate(files):
            if i == 0:
                window = self.windows[0]
            else:
                window = MainWindow(self)
            window.load_file(file)
            window.present()
            self.windows.append(window)

    def _set_theme(self, settings, *_pspec):
        # Third party themes cause issues with Apostrophe custom stylesheets
        # If the user has a third party theme selected, we just change it to
        # Adwaita to prevent those issues

        # TODO: GTK4 - remove this
 
        theme_name = settings.get_property("gtk-theme-name")
        icon_theme_name = settings.get_property("gtk-icon-theme-name")

        if (theme_name not in ["Adwaita",
                               "HighContrast",
                               "HighContrastInverse"]):
            settings.set_property("gtk-theme-name", "Adwaita")


    def _set_color_scheme(self, *args, **kwargs):
        # TODO: GTK4 - remove this
        theme = Theme.get_current()

        style_manager = Handy.StyleManager.get_default()
        dark = style_manager.get_dark()
        sepia = theme.name == "sepia"
        hc = style_manager.get_high_contrast()

        if not self.windows:
            return

        if sepia:
            if hc:
                self.style_provider.load_from_file(theme.gtk_css_hc)
            else:
                self.style_provider.load_from_file(theme.gtk_css)
        elif dark:
            if hc:
                self.style_provider.load_from_file(Theme.get_for_name("dark").gtk_css_hc)
            else:
                self.style_provider.load_from_file(Theme.get_for_name("dark").gtk_css)
        else:
            if hc:
                self.style_provider.load_from_file(Theme.get_for_name("light").gtk_css_hc)
            else:
                self.style_provider.load_from_file(Theme.get_for_name("light").gtk_css)


    def on_settings_changed(self, settings, key):
        # TODO: change this ffs
        if not self.windows:
            return
        if key == "color-scheme":
            self._set_color_scheme()
        elif key == "input-format":
            for window in self.windows:
                window.reload_preview()
        elif key == "sync-scroll":
            for window in self.windows:
                window.reload_preview(reshow=True)
        elif key == "stat-default":
            for window in self.windows:
                window.update_default_stat()

    def on_new_window(self, _action, _value):
        window = MainWindow(self)
        window.present()
        self.windows.append(window)

    def on_preferences(self, _action, _value):
        preferences_dialog = ApostrophePreferencesDialog()
        preferences_dialog.set_transient_for(self.get_active_window())
        preferences_dialog.show()

    def on_shortcuts(self, _action, _param):
        builder = Gtk.Builder()
        builder.add_from_resource(
            "/org/gnome/gitlab/somas/Apostrophe/ui/Shortcuts.ui")
        builder.get_object("shortcuts").set_transient_for(self.get_active_window())
        builder.get_object("shortcuts").show()

    def on_about(self, _action, _param):
        # TODO: what about non-csd
        builder = Gtk.Builder()
        builder.add_from_resource(
            "/org/gnome/gitlab/somas/Apostrophe/About.ui")
        about_dialog = builder.get_object("AboutDialog")
        about_dialog.set_transient_for(self.get_active_window())

        about_dialog.present()

    def on_quit(self, _action, _param):
        quit = True
        for window in self.windows:
            window.present()
            if window.on_delete_called(self):
                quit = False
        if quit:
            self.quit()

    def on_stat_default(self, action, value):
        action.set_state(value)
        self.settings.set_string("stat-default", value.get_string())



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from gi.repository import Pango
from gi.repository import Gtk, GLib
import re
from multiprocessing import Pipe, Process

import gi

from apostrophe import helpers, markup_regex
from apostrophe.markup_regex import STRIKETHROUGH, BOLD_ITALIC,\
    BOLD, ITALIC_ASTERISK, ITALIC_UNDERSCORE, IMAGE, LINK, LINK_ALT,\
    HORIZONTAL_RULE, LIST, ORDERED_LIST, BLOCK_QUOTE, HEADER, HEADER_UNDER,\
    TABLE, MATH, CODE

gi.require_version('Gtk', '3.0')


class MarkupHandler:
    TAG_NAME_ITALIC = 'italic'
    TAG_NAME_BOLD = 'bold'
    TAG_NAME_BOLD_ITALIC = 'bold_italic'
    TAG_NAME_STRIKETHROUGH = 'strikethrough'
    TAG_NAME_CENTER = 'center'
    TAG_NAME_WRAP_NONE = 'wrap_none'
    TAG_NAME_PLAIN_TEXT = 'plain_text'
    TAG_NAME_GRAY_TEXT = 'gray_text'
    TAG_NAME_CODE_TEXT = 'code_text'
    TAG_NAME_CODE_BLOCK = 'code_block'
    TAG_NAME_UNFOCUSED_TEXT = 'unfocused_text'
    TAG_NAME_MARGIN_INDENT = 'margin_indent'

    def __init__(self, text_view):
        self.text_view = text_view
        self.text_buffer = text_view.get_buffer()
        self.marked_up_text = None

        # Tags.
        buffer = self.text_buffer

        self.tag_italic = buffer.create_tag(self.TAG_NAME_ITALIC,
                                            weight=Pango.Weight.NORMAL,
                                            style=Pango.Style.ITALIC)

        self.tag_bold = buffer.create_tag(self.TAG_NAME_BOLD,
                                          weight=Pango.Weight.BOLD,
                                          style=Pango.Style.NORMAL)

        self.tag_bold_italic = buffer.create_tag(self.TAG_NAME_BOLD_ITALIC,
                                                 weight=Pango.Weight.BOLD,
                                                 style=Pango.Style.ITALIC)

        self.tag_strikethrough = buffer.create_tag(self.TAG_NAME_STRIKETHROUGH,
                                                   strikethrough=True)

        self.tag_center = buffer.create_tag(self.TAG_NAME_CENTER,
                                            justification=Gtk.Justification.CENTER)

        self.tag_wrap_none = buffer.create_tag(self.TAG_NAME_WRAP_NONE,
                                               wrap_mode=Gtk.WrapMode.NONE,
                                               pixels_above_lines=0,
                                               pixels_below_lines=0)

        self.tag_plain_text = buffer.create_tag(self.TAG_NAME_PLAIN_TEXT,
                                                weight=Pango.Weight.NORMAL,
                                                style=Pango.Style.NORMAL,
                                                strikethrough=False,
                                                justification=Gtk.Justification.LEFT)

        self.tag_gray_text = buffer.create_tag(self.TAG_NAME_GRAY_TEXT,
                                               foreground='gray',
                                               weight=Pango.Weight.NORMAL,
                                               style=Pango.Style.NORMAL)

        self.tag_code_text = buffer.create_tag(self.TAG_NAME_CODE_TEXT,
                                               weight=Pango.Weight.NORMAL,
                                               style=Pango.Style.NORMAL,
                                               strikethrough=False)

        self.tag_code_block = buffer.create_tag(self.TAG_NAME_CODE_BLOCK,
                                                weight=Pango.Weight.NORMAL,
                                                style=Pango.Style.NORMAL,
                                                strikethrough=False,
                                                indent=self.get_margin_indent(0, 1)[1])

        self.tags_markup = {
            self.TAG_NAME_ITALIC: lambda args: self.tag_italic,
            self.TAG_NAME_BOLD: lambda args: self.tag_bold,
            self.TAG_NAME_BOLD_ITALIC: lambda args: self.tag_bold_italic,
            self.TAG_NAME_STRIKETHROUGH: lambda args: self.tag_strikethrough,
            self.TAG_NAME_CENTER: lambda args: self.tag_center,
            self.TAG_NAME_WRAP_NONE: lambda args: self.tag_wrap_none,
            self.TAG_NAME_PLAIN_TEXT: lambda args: self.tag_plain_text,
            self.TAG_NAME_GRAY_TEXT: lambda args: self.tag_gray_text,
            self.TAG_NAME_CODE_TEXT: lambda args: self.tag_code_text,
            self.TAG_NAME_CODE_BLOCK: lambda args: self.tag_code_block,
            self.TAG_NAME_MARGIN_INDENT: lambda args: self.get_margin_indent_tag(
                *args)
        }

        # Focus mode.
        self.tag_unfocused_text = buffer.create_tag(self.TAG_NAME_UNFOCUSED_TEXT,
                                                    foreground='gray',
                                                    weight=Pango.Weight.NORMAL,
                                                    style=Pango.Style.NORMAL)

        # Margin and indents.
        # A baseline margin is set to allow negative offsets for formatting
        # headers, lists, etc.
        self.tags_margins_indents = {}
        self.baseline_margin = 0
        self.char_width = 0
        self.update_margins_indents()

        # Style.
        self.on_style_updated()

        # Worker process to handle parsing.
        self.parsing = False
        self.apply_pending = False
        self.parent_conn, child_conn = Pipe()
        Process(target=self.parse, args=(child_conn,), daemon=True).start()
        GLib.io_add_watch(
            self.parent_conn.fileno(),
            GLib.PRIORITY_DEFAULT,
            GLib.IO_IN,
            self.on_parsed)

    def on_style_updated(self, *_):
        style_context = self.text_view.get_style_context()
        (found, color) = style_context.lookup_color('code_bg_color')
        if not found:
            (_, color) = style_context.lookup_color('background_color')
        self.tag_code_text.set_property("background", color.to_string())
        self.tag_code_block.set_property(
            "paragraph-background", color.to_string())

    def apply(self):
        """Applies markup, parsing it in a worker process
        if the text has changed.

        In case parsing is already running, it will re-apply once it finishes.
        This ensure that the pipe doesn't fill (and block) if multiple requests
        are made in quick succession."""

        if not self.parsing:
            self.parsing = True
            self.apply_pending = False

            text = self.text_buffer.get_slice(
                self.text_buffer.get_start_iter(),
                self.text_buffer.get_end_iter(),
                False)
            if text != self.marked_up_text:
                self.parent_conn.send(text)
            else:
                self.do_apply(text)
        else:
            self.apply_pending = True

    def parse(self, child_conn):
        """Parses markup in a worker process."""

        while True:
            while True:
                try:
                    text = child_conn.recv()
                    if not child_conn.poll():
                        break
                except EOFError:
                    child_conn.close()
                    return

            # List of tuples in the form (tag_name, tag_args, tag_start,
            # tag_end).
            result = []

            # Find:
            # - "_italic_" (italic)
            # - "**bold**" (bold)
            # - "***bolditalic***" (bold/italic)
            # - "~~strikethrough~~" (strikethrough)
            # - "`code`" (colorize)
            # - "$math$" (colorize)
            # - "---" table (wrap/pixels)
            regexps = (
                (ITALIC_ASTERISK, self.TAG_NAME_ITALIC),
                (ITALIC_UNDERSCORE, self.TAG_NAME_ITALIC),
                (BOLD, self.TAG_NAME_BOLD),
                (BOLD_ITALIC, self.TAG_NAME_BOLD_ITALIC),
                (STRIKETHROUGH, self.TAG_NAME_STRIKETHROUGH),
                (CODE, self.TAG_NAME_CODE_TEXT),
                (MATH, self.TAG_NAME_CODE_TEXT),
                (TABLE, self.TAG_NAME_WRAP_NONE)
            )
            for regexp, tag_name in regexps:
                matches = re.finditer(regexp, text)
                for match in matches:
                    result.append((tag_name, (), match.start(), match.end()))

            # Find:
            # - "[description](url)" (gray out)
            # - "![description](image_url)" (gray out)
            regexps = (
                (LINK, self.TAG_NAME_GRAY_TEXT),
                (IMAGE, self.TAG_NAME_GRAY_TEXT)
            )
            for regexp, tag_name in regexps:
                matches = re.finditer(regexp, text)
                for match in matches:
                    result.append(
                        (tag_name, (), match.start(), match.start("text")))
                    result.append(
                        (tag_name, (), match.end("text"), match.end()))

            # Find "<url>" links (gray out).
            matches = re.finditer(LINK_ALT, text)
            for match in matches:
                result.append((
                    self.TAG_NAME_GRAY_TEXT,
                    (), match.start("text"),
                    match.end("text")))

            # Find "---" horizontal rule (center).
            matches = re.finditer(HORIZONTAL_RULE, text)
            for match in matches:
                result.append((
                    self.TAG_NAME_CENTER,
                    (), match.start("symbols"),
                    match.end("symbols")))

            # Find "* list" (offset).
            matches = re.finditer(LIST, text)
            for match in matches:
                # Lists use character+space (eg. "* ").
                length = 2
                nest = len(match.group("indent").replace("    ", "\t"))
                margin = -length - 2 * nest
                indent = -length - 2 * length * nest
                result.append((
                    self.TAG_NAME_MARGIN_INDENT,
                    (margin, indent),
                    match.start("content"),
                    match.end("content")
                ))

            # Find "1. ordered list" (offset).
            matches = re.finditer(ORDERED_LIST, text)
            for match in matches:
                # Ordered lists use numbers/letters+dot/parens+space
                # (eg. "123. ").
                length = len(match.group("prefix")) + 1
                nest = len(match.group("indent").replace("    ", "\t"))
                margin = -length - 2 * nest
                indent = -length - 2 * length * nest
                result.append((
                    self.TAG_NAME_MARGIN_INDENT,
                    (margin, indent),
                    match.start("content"),
                    match.end("content")
                ))

            # Find "> blockquote" (offset).
            matches = re.finditer(BLOCK_QUOTE, text)
            for match in matches:
                result.append((self.TAG_NAME_MARGIN_INDENT,
                               (2, -2), match.start(), match.end()))

            # Find "# Header" (offset+bold).
            matches = re.finditer(HEADER, text)
            for match in matches:
                margin = -len(match.group("level")) - 1
                result.append((
                    self.TAG_NAME_MARGIN_INDENT, (margin, 0),
                    match.start(), match.end()))
                result.append(
                    (self.TAG_NAME_BOLD, (), match.start(), match.end()))

            # Find "=======" header underline (bold).
            matches = re.finditer(HEADER_UNDER, text)
            for match in matches:
                result.append(
                    (self.TAG_NAME_BOLD, (), match.start(), match.end()))

            # Find "```" code block tag (offset + colorize paragraph).
            matches = re.finditer(markup_regex.CODE_BLOCK, text)
            for match in matches:
                result.append((
                    self.TAG_NAME_CODE_BLOCK, (),
                    match.start("block"), match.end("block")))

            # Send parsed data back.
            child_conn.send((text, result))

    def on_parsed(self, _source, _condition):
        """Reads the parsing result from the pipe
        and triggers any pending apply."""

        self.parsing = False
        if self.apply_pending:
            self.apply()  # self.apply clears the apply pending flag.

        try:
            if self.parent_conn.poll():
                self.do_apply(*self.parent_conn.recv())
            return True
        except EOFError:
            return False

    def do_apply(self, original_text, result=[]):
        """Applies the result of parsing if the current text
        matches the original text."""

        buffer = self.text_buffer
        start = buffer.get_start_iter()
        end = buffer.get_end_iter()
        text = self.text_buffer.get_slice(start, end, False)

        # Apply markup tags.
        if text == original_text and text != self.marked_up_text:
            buffer.remove_tag(self.tag_italic, start, end)
            buffer.remove_tag(self.tag_bold, start, end)
            buffer.remove_tag(self.tag_bold_italic, start, end)
            buffer.remove_tag(self.tag_strikethrough, start, end)
            buffer.remove_tag(self.tag_center, start, end)
            buffer.remove_tag(self.tag_plain_text, start, end)
            buffer.remove_tag(self.tag_gray_text, start, end)
            buffer.remove_tag(self.tag_code_text, start, end)
            buffer.remove_tag(self.tag_code_block, start, end)
            buffer.remove_tag(self.tag_wrap_none, start, end)
            for tag in self.tags_margins_indents.values():
                buffer.remove_tag(tag, start, end)

            for tag_name, tag_args, tag_start, tag_end in result:
                buffer.apply_tag(
                    self.tags_markup[tag_name](tag_args),
                    buffer.get_iter_at_offset(tag_start),
                    buffer.get_iter_at_offset(tag_end))

        # Apply focus mode tag (grey out before/after current sentence).
        buffer.remove_tag(self.tag_unfocused_text, start, end)
        if self.text_view.focus_mode:
            cursor_iter = buffer.get_iter_at_mark(buffer.get_insert())
            start_sentence = cursor_iter.copy()
            if not start_sentence.starts_sentence():
                start_sentence.backward_sentence_start()
            end_sentence = cursor_iter.copy()
            if not end_sentence.ends_sentence():
                end_sentence.forward_sentence_end()
            buffer.apply_tag(self.tag_unfocused_text, start, start_sentence)
            buffer.apply_tag(self.tag_unfocused_text, end_sentence, end)

    # Margin and indent are cumulative. They differ in two ways:
    # * Margin is always in the beginning,
    # which means it effectively only affects the first line
    # of multi-line text. Indent is applied to every line.
    # * Margin level can be negative, as a baseline margin exists
    # from which it can be subtracted.
    # Indent is always positive, or 0.
    def get_margin_indent_tag(self, margin_level, indent_level):
        level = (margin_level, indent_level)
        if level not in self.tags_margins_indents:
            margin, indent = self.get_margin_indent(margin_level, indent_level)
            tag = self.text_buffer.create_tag(
                "margin_indent_{}_{}".format(margin_level, indent_level),
                left_margin=margin, indent=indent)
            self.tags_margins_indents[level] = tag
            return tag
        else:
            return self.tags_margins_indents[level]

    def get_margin_indent(self, margin_level, indent_level,
                          baseline_margin=None, char_width=None):
        if baseline_margin is None:
            baseline_margin = self.text_view.props.left_margin
        if char_width is None:
            char_width = helpers.get_char_width(self.text_view)
        margin = max(baseline_margin + char_width * margin_level, 0)
        indent = char_width * indent_level
        return margin, indent

    def update_margins_indents(self):
        baseline_margin = self.text_view.props.left_margin
        char_width = helpers.get_char_width(self.text_view)

        # Bail out if neither the baseline margin nor character width change
        if baseline_margin == self.baseline_margin and char_width == self.char_width:
            return
        self.baseline_margin = baseline_margin
        self.char_width = char_width

        # Adjust tab size
        tab_array = Pango.TabArray.new(1, True)
        tab_array.set_tab(0, Pango.TabAlign.LEFT, 4 * char_width)
        self.text_view.set_tabs(tab_array)

        # Adjust margins and indents
        for level, tag in self.tags_margins_indents.items():
            margin, indent = self.get_margin_indent(
                *level, baseline_margin, char_width)
            tag.set_properties(left_margin=margin, indent=indent)

    def stop(self, *_):
        self.parent_conn.close()



import re
from multiprocessing import Process, Pipe

from gi.repository import GLib

from apostrophe.markup_regex import ITALIC_ASTERISK, ITALIC_UNDERSCORE,\
    BOLD_ITALIC, BOLD, STRIKETHROUGH, IMAGE, LINK, LINK_ALT, HORIZONTAL_RULE,\
    LIST, MATH, TABLE, CODE_BLOCK, HEADER_UNDER, HEADER, BLOCK_QUOTE,\
    ORDERED_LIST, FOOTNOTE_ID, FOOTNOTE


class StatsCounter:
    """Counts characters, words, sentences
    and read time using a worker process."""

    # Regexp that matches any character,
    # except for newlines and subsequent spaces.
    CHARACTERS = re.compile(r"[^\s]|(?:[^\S\n](?!\s))")

    # Regexp that matches Asian letters, general symbols and hieroglyphs,
    # as well as sequences of word characters optionally containing non-word
    # characters in-between.
    WORDS = re.compile(r"[\u3040-\uffff]|(?:\w+\S?\w*)+", re.UNICODE)

    # Regexp that matches sentence-ending punctuation characters, ie.
    # full stop, question mark, exclamation mark, paragraph, and variants.
    SENTENCES = re.compile(r"[^\n][.。।෴۔።?՞;⸮؟？፧꘏⳺⳻⁇﹖⁈⁉‽!﹗！՜߹႟᥄\n]+")

    # Regexp that matches paragraphs, ie.
    # anything separated by at least 2 newlines.
    PARAGRAPHS = re.compile(r"[^\n]+(\n{2,}|$)")

    # List of regexp whose matches should be replaced by their "text" group.
    # Order is important.
    MARKUP_REGEXP_REPLACE = (
        BOLD_ITALIC, ITALIC_ASTERISK, ITALIC_UNDERSCORE, BOLD, STRIKETHROUGH,
        IMAGE, LINK, LINK_ALT, LIST, ORDERED_LIST, BLOCK_QUOTE, HEADER,
        HEADER_UNDER, CODE_BLOCK, TABLE, MATH, FOOTNOTE_ID, FOOTNOTE
    )

    # List of regexp whose matches should be removed. Order is important.
    MARKUP_REGEXP_REMOVE = (
        HORIZONTAL_RULE,
    )

    def __init__(self, callback):
        super().__init__()

        # Worker process to handle counting.
        self.counting = False
        self.count_pending_text = None
        self.parent_conn, child_conn = Pipe()
        Process(target=self.do_count, args=(child_conn,), daemon=True).start()
        GLib.io_add_watch(
            self.parent_conn.fileno(), GLib.PRIORITY_LOW, GLib.IO_IN,
            self.on_counted, callback)

    def count(self, text):
        """Count stats for text.

        In case counting is already running, it will re-count once it finishes.
        This ensure that the pipe doesn't fill (and block) if multiple requests
        are made in quick succession."""

        if not self.counting:
            self.counting = True
            self.count_pending_text = None
            self.parent_conn.send(text)
        else:
            self.count_pending_text = text

    def do_count(self, child_conn):
        """Counts stats in a worker process.

        The result is in the format
        (characters, words, sentences, (hours, minutes, seconds))"""

        while True:
            while True:
                try:
                    text = child_conn.recv()
                    if not child_conn.poll():
                        break
                except EOFError:
                    child_conn.close()
                    return

            for regexp in self.MARKUP_REGEXP_REPLACE:
                text = re.sub(regexp, r"\g<text>", text)
            for regexp in self.MARKUP_REGEXP_REMOVE:
                text = re.sub(regexp, "", text)

            character_count = len(re.findall(self.CHARACTERS, text))

            word_count = len(re.findall(self.WORDS, text))

            sentence_count = len(re.findall(self.SENTENCES, text))

            paragraph_count = len(re.findall(self.PARAGRAPHS, text))

            read_m, read_s = divmod(word_count / 200 * 60, 60)
            read_h, read_m = divmod(read_m, 60)
            read_time = (int(read_h), int(read_m), int(read_s))

            child_conn.send(
                (character_count, word_count, sentence_count,
                    paragraph_count, read_time))

    def on_counted(self, _source, _condition, callback):
        """Reads the counting result from the pipe
        and triggers any pending count."""

        self.counting = False
        if self.count_pending_text is not None:
            # self.count clears the pending text.
            self.count(self.count_pending_text)

        try:
            if self.parent_conn.poll():
                callback(self.parent_conn.recv())
            return True
        except EOFError:
            return False

    def stop(self):
        """Stops the worker process.
        StatsCounter shouldn't be used after this."""

        self.parent_conn.close()



# Copyright (C) 2021, Manuel Genovés <manuel.genoves@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, GLib, Gio, Handy


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/PreviewWindow.ui')
class PreviewWindow(Handy.ApplicationWindow):

    __gtype_name__ = "ApostrophePreviewWindow"

    preview_box = Gtk.Template.Child()

    def __init__(self):
        super().__init__(application=Gio.Application.get_default(),
                         title="Preview")



# BEGIN LICENSE
# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
# Copyright (C) 2021, Adrien Plazas <kekun.plazas@laposte.net>
# Copyright (C) 2021, Alexander Mikhaylenko <exalm7659@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from apostrophe.settings import Settings
from gi.repository import Gtk, GObject
from enum import IntEnum
import enum

from gettext import gettext as _
import gi
gi.require_version('Gtk', '3.0')


class PreviewLayout(IntEnum):

    # preview modes
    FULL_WIDTH = 0
    HALF_WIDTH = 1
    HALF_HEIGHT = 2
    WINDOWED = 3

    def get_text(self):
        if self == PreviewLayout.FULL_WIDTH:
            return _("Full-Width")
        elif self == PreviewLayout.HALF_WIDTH:
            return _("Half-Width")
        elif self == PreviewLayout.HALF_HEIGHT:
            return _("Half-Height")
        elif self == PreviewLayout.WINDOWED:
            return _("Windowed")
        else:
            raise ValueError("Unknown preview mode")

    def get_icon(self):
        if self == PreviewLayout.FULL_WIDTH:
            return "preview-layout-full-width-symbolic"
        elif self == PreviewLayout.HALF_WIDTH:
            return "preview-layout-half-width-symbolic"
        elif self == PreviewLayout.HALF_HEIGHT:
            return "preview-layout-half-height-symbolic"
        elif self == PreviewLayout.WINDOWED:
            return "preview-layout-windowed-symbolic"
        else:
            raise ValueError("Unknown preview mode")


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/PreviewLayoutSwitcherItem.ui')
class PreviewLayoutSwitcherItem(Gtk.ListBoxRow):
    __gtype_name__ = "PreviewLayoutSwitcherItem"

    icon = Gtk.Template.Child()
    title = Gtk.Template.Child()
    checkmark = Gtk.Template.Child()
    selected = GObject.property(type=bool, default=False)

    def __init__(self, _layout):
        super().__init__()
        self.layout = _layout
        self.icon.set_from_icon_name(self.layout.get_icon(), 4)
        self.title.set_label(self.layout.get_text())

        self.connect("notify::selected", self.on_selected)

    def on_selected(self, *args, **kwargs):
        self.checkmark.props.opacity = 1 if self.selected else 0


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/PreviewLayoutSwitcher.ui')
class PreviewLayoutSwitcher(Gtk.Box):
    __gtype_name__ = "PreviewLayoutSwitcher"

    list_box = Gtk.Template.Child()
    layout_image = Gtk.Template.Child()
    layout_popover = Gtk.Template.Child()
    preview_switcher_toggle = Gtk.Template.Child()
    items = []

    preview_layout = GObject.property(type=int)
    settings = Settings.new()

    def __init__(self):
        super().__init__()

        for i, layout in enumerate(PreviewLayout):
            item = PreviewLayoutSwitcherItem(layout)
            self.list_box.add(item)
            self.items.append(item)

        self.update_ui()
        self.connect("notify::preview-layout", self.update_ui)

    @Gtk.Template.Callback()
    def update_ui(self, *args, **kwargs):
        self.layout_image.set_from_icon_name(PreviewLayout(self.preview_layout).get_icon(), 4)

        for item in self.items:
            item.selected = item.layout == self.preview_layout

        item = self.items[self.preview_layout]
        self.list_box.select_row(item)

    @Gtk.Template.Callback()
    def on_row_activated(self, _listbox, row):
        self.preview_layout = row.layout
        self.layout_popover.popdown()

    # TODO: remove
    @Gtk.Template.Callback()
    def on_menu_state_changed(self, _listbox, row):
        pass



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import re
import os
import telnetlib
from gettext import gettext as _
from urllib.parse import unquote

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("WebKit2", "4.0")
from gi.repository import Gtk, Gdk, GdkPixbuf, GLib
from gi.repository import WebKit2
from apostrophe import latex_to_PNG, markup_regex
from apostrophe.settings import Settings


class DictAccessor:
    reEndResponse = re.compile(br"^[2-5][0-58][0-9] .*\r\n$", re.DOTALL +
                               re.MULTILINE)
    reDefinition = re.compile(br"^151(.*?)^\.", re.DOTALL + re.MULTILINE)

    def __init__(self, host="dict.dict.org", port=2628, timeout=60):
        self.telnet = telnetlib.Telnet(host, port)
        self.timeout = timeout
        self.login_response = self.telnet.expect([self.reEndResponse],
                                                 self.timeout)[2]

    def run_command(self, cmd):
        self.telnet.write(cmd.encode("utf-8") + b"\r\n")
        return self.telnet.expect([self.reEndResponse], self.timeout)[2]

    def get_matches(self, database, strategy, word):
        if database in ["", "all"]:
            d = "*"
        else:
            d = database
        if strategy in ["", "default"]:
            s = "."
        else:
            s = strategy
        w = word.replace("\"", r"\\\"")
        tsplit = self.run_command(
            "MATCH {} {} \"{}\"".format(
                d, s, w)).splitlines()
        mlist = list()
        if tsplit[-1].startswith(b"250 ok") and tsplit[0].startswith(b"1"):
            mlines = tsplit[1:-2]
            for line in mlines:
                lsplit = line.strip().split()
                db = lsplit[0]
                word = unquote(" ".join(lsplit[1:]))
                mlist.append((db, word))
        return mlist

    def get_definition(self, database, word):
        if database in ["", "all"]:
            d = "*"
        else:
            d = database
        w = word.replace("\"", r"\\\"")
        dsplit = self.run_command(
            "DEFINE {} \"{}\"".format(
                d, w)).splitlines(True)

        dlist = list()
        if dsplit[-1].startswith(b"250 ok") and dsplit[0].startswith(b"1"):
            dlines = dsplit[1:-1]
            dtext = b"".join(dlines)
            dlist = [dtext]
        return dlist

    def close(self):
        t = self.run_command("QUIT")
        self.telnet.close()
        return t

    def parse_wordnet(self, response):
        # consisting of group (n,v,adj,adv)
        # number, description, examples, synonyms, antonyms

        lines = response.splitlines()
        lines = lines[2:]
        lines = " ".join(lines)
        lines = re.sub(r"\s+", " ", lines).strip()
        lines = re.split(r"( adv | adj | n | v |^adv |^adj |^n |^v )", lines)
        res = []
        act_res = {"defs": [], "class": "none", "num": "None"}
        for l in lines:
            l = l.strip()
            if not l:
                continue
            if l in ["adv", "adj", "n", "v"]:
                if act_res:
                    res.append(act_res.copy())
                act_res = {"defs": [], "class": l}
            else:
                ll = re.split(r"(?: |^)(\d): ", l)
                act_def = {}
                for lll in ll:
                    if lll.strip().isdigit() or not lll.strip():
                        if "description" in act_def and act_def["description"]:
                            act_res["defs"].append(act_def.copy())
                        act_def = {"num": lll}
                        continue
                    a = re.findall(r"(\[(syn|ant): (.+?)\] ??)+", lll)
                    for n in a:
                        if n[1] == "syn":
                            act_def["syn"] = re.findall(r"{(.*?)}.*?", n[2])
                        else:
                            act_def["ant"] = re.findall(r"{(.*?)}.*?", n[2])
                    tbr = re.search(r"\[.+\]", lll)
                    if tbr:
                        lll = lll[:tbr.start()]
                    lll = lll.split(";")
                    act_def["examples"] = []
                    act_def["description"] = []
                    for llll in lll:
                        llll = llll.strip()
                        if llll.strip().startswith("\""):
                            act_def["examples"].append(llll)
                        else:
                            act_def["description"].append(llll)
                if act_def and "description" in act_def:
                    act_res["defs"].append(act_def.copy())

        res.append(act_res.copy())
        return res


def get_dictionary(term):
    da = DictAccessor()
    output = da.get_definition("wn", term)
    if output:
        output = output[0]
    else:
        return None
    return da.parse_wordnet(output.decode(encoding="UTF-8"))


class InlinePreview:
    WIDTH = 400
    HEIGHT = 300

    def __init__(self, text_view):
        self.settings = Settings.new()

        self.text_view = text_view
        self.text_view.connect("button-press-event",
                               self.on_button_press_event)
        self.text_buffer = text_view.get_buffer()
        self.cursor_mark = self.text_buffer.create_mark(
            "click",
            self.text_buffer.get_iter_at_mark(self.text_buffer.get_insert()))

        self.latex_converter = latex_to_PNG.LatexToPNG()
        self.characters_per_line = self.settings.get_int("characters-per-line")

        self.popover = Gtk.Popover.new(self.text_view)
        self.popover.get_style_context().add_class("quick-preview-popup")
        self.popover.set_modal(True)

        self.preview_fns = {
            markup_regex.MATH: self.get_view_for_math,
            markup_regex.IMAGE: self.get_view_for_image,
            markup_regex.LINK: self.get_view_for_link,
            markup_regex.LINK_ALT: self.get_view_for_link,
            markup_regex.FOOTNOTE_ID: self.get_view_for_footnote,
            re.compile(r"(?P<text>\w+)"): self.get_view_for_lexikon
        }

    def on_button_press_event(self, _text_view, event):
        if event.button == 1 and event.state & Gdk.ModifierType.CONTROL_MASK:
            x, y = self.text_view.window_to_buffer_coords(2, int(event.x),
                                                          int(event.y))
            self.text_buffer.move_mark(
                self.cursor_mark,
                self.text_view.get_iter_at_location(x, y).iter)
            self.open_popover(self.text_view)

    def get_view_for_math(self, match):
        success, result = self.latex_converter.generatepng(match.group("text"))
        if success:
            return Gtk.Image.new_from_file(result)
        else:
            error = _("Formula looks incorrect:")
            error += "\n\n“{}”".format(result)
            return Gtk.Label(label=error)

    def get_view_for_image(self, match):
        path = match.group("url")

        if path.startswith(("https://", "http://", "www.")):
            return self.get_view_for_link(match)
        if path.startswith(("file://")):
            path = path[7:]
        if not path.startswith(("/", "file://")):
            path = os.path.join(
                self.settings.get_string("open-file-path"), path)
        path = unquote(path)

        return Gtk.Image.new_from_pixbuf(
            GdkPixbuf.Pixbuf.new_from_file_at_size(path, self.WIDTH,
                                                   self.HEIGHT))

    def get_view_for_link(self, match):
        url = match.group("url")
        web_view = WebKit2.WebView(zoom_level=0.3125)  # ~1280x960
        web_view.set_size_request(self.WIDTH, self.HEIGHT)
        if GLib.uri_parse_scheme(url) is None:
            url = "http://{}".format(url)
        web_view.load_uri(url)
        return web_view

    def get_view_for_footnote(self, match):
        footnote_id = match.group("id")
        fn_matches = re.finditer(
            markup_regex.FOOTNOTE,
            self.text_buffer.props.text)
        for fn_match in fn_matches:
            if fn_match.group("id") == footnote_id:
                if fn_match:
                    footnote = re.sub("\n[\t ]+", "\n", fn_match.group("text"))
                else:
                    footnote = _("No matching footnote found")
                label = Gtk.Label(label=footnote)
                label.set_max_width_chars(self.characters_per_line)
                label.set_line_wrap(True)
                return label
        return None

    def get_view_for_lexikon(self, match):
        term = match.group("text")
        lexikon_dict = get_dictionary(term)
        if lexikon_dict:
            grid = Gtk.Grid.new()
            grid.get_style_context().add_class("lexikon")
            grid.set_row_spacing(2)
            grid.set_column_spacing(4)
            i = 0
            for entry in lexikon_dict:
                if not entry["defs"]:
                    continue
                elif entry["class"].startswith("n"):
                    word_type = _("noun")
                elif entry["class"].startswith("v"):
                    word_type = _("verb")
                elif entry["class"].startswith("adj"):
                    word_type = _("adjective")
                elif entry["class"].startswith("adv"):
                    word_type = _("adverb")
                else:
                    continue

                vocab_label = Gtk.Label.new(term + " ~ " + word_type)
                vocab_label.get_style_context().add_class("header")
                if i == 0:
                    vocab_label.get_style_context().add_class("first")
                vocab_label.set_halign(Gtk.Align.START)
                vocab_label.set_justify(Gtk.Justification.LEFT)
                grid.attach(vocab_label, 0, i, 3, 1)

                for definition in entry["defs"]:
                    i = i + 1
                    num_label = Gtk.Label.new(definition["num"] + ".")
                    num_label.get_style_context().add_class("number")
                    num_label.set_valign(Gtk.Align.START)
                    grid.attach(num_label, 0, i, 1, 1)

                    def_label = Gtk.Label(
                        label=" ".join(
                            definition["description"]))
                    def_label.get_style_context().add_class("description")
                    def_label.set_halign(Gtk.Align.START)
                    def_label.set_max_width_chars(self.characters_per_line)
                    def_label.set_line_wrap(True)
                    def_label.set_justify(Gtk.Justification.FILL)
                    grid.attach(def_label, 1, i, 1, 1)
                i = i + 1
            if i > 0:
                return grid
        return None

    def open_popover(self, _editor, _data=None):
        start_iter = self.text_buffer.get_iter_at_mark(self.cursor_mark)
        line_offset = start_iter.get_line_offset()
        end_iter = start_iter.copy()
        start_iter.set_line_offset(0)
        end_iter.forward_to_line_end()
        text = self.text_buffer.get_text(start_iter, end_iter, False)

        for regex, get_view_fn in self.preview_fns.items():
            matches = re.finditer(regex, text)
            for match in matches:
                if match.start() <= line_offset <= match.end():
                    prev_view = self.popover.get_child()
                    if prev_view:
                        prev_view.destroy()
                    view = get_view_fn(match)
                    if view:
                        view.show_all()
                        self.popover.add(view)
                        rect = self.text_view.get_iter_location(
                            self.text_buffer.get_iter_at_mark(self.cursor_mark))
                        rect.x, rect.y = self.text_view.buffer_to_window_coords(
                            Gtk.TextWindowType.TEXT, rect.x, rect.y)
                        self.popover.set_pointing_to(rect)
                        GLib.idle_add(self.popover.popup)
                    return



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2021, Christian Hergert <chergert@redhat.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE
"""Open/recents popover
"""

from os import close
import gi

from gettext import gettext as _

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio, GLib, GObject, Handy
from .settings import Settings

class RecentItem(GObject.Object):
    def __init__(self, name, path, uri, **kwargs):
        super().__init__(**kwargs)
        self.name: str = name
        self.path: str = path
        self.uri: str = uri

@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/Recents.ui')
class ApostropheOpenPopover(Gtk.Popover):

    __gtype_name__ = "ApostropheOpenPopover"

    """Open/recents popover
    """

    list_box = Gtk.Template.Child()
    stack = Gtk.Template.Child()
    empty = Gtk.Template.Child()
    recent = Gtk.Template.Child()


    model = Gio.ListStore.new(RecentItem)
    recents_manager = Gtk.RecentManager.get_default()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.list_box.bind_model(self.model, self.create_row)

        self.on_manager_changed()
        self.recents_manager.connect("changed", self.on_manager_changed)

    def create_row(self, item, **args):
        row = Handy.ActionRow.new()
        row.item = item
        row.set_title(item.name)
        row.set_subtitle(item.path)

        delete_button = Gtk.Button.new_from_icon_name("window-close-symbolic", Gtk.IconSize.BUTTON)
        delete_button.get_style_context().add_class("flat")
        delete_button.get_style_context().add_class("circular")
        delete_button.set_valign(Gtk.Align.CENTER)
        delete_button.set_visible(True)
        delete_button.connect("clicked", self.on_delete_click, item)

        row.add(delete_button)
        row.set_activatable(True)
        row.set_action_name("win.open_file")
        row.set_action_target_value(GLib.Variant.new_string(item.uri))

        return row

    def on_manager_changed(self, *args, **kwargs):
        self.model.remove_all()
        for item in self.recents_manager.get_items():
            self.model.append(RecentItem(item.get_display_name(), item.get_uri_display(), item.get_uri()))

        self.stack.set_visible_child(self.recent if self.model else self.empty)

    @Gtk.Template.Callback()
    def on_search_entry_changed_cb(self, search_entry):
        # TODO: implement nice filters in GTK4
        recents_list = self.recents_manager.get_items()
        filtered_list = filter(lambda item: search_entry.get_text() in item.get_display_name(), recents_list)

        self.model.remove_all()
        for item in filtered_list:
            self.model.append(RecentItem(item.get_display_name(), item.get_uri_display(), item.get_uri()))

    @Gtk.Template.Callback()
    def on_search_entry_activate_cb(self, *arg, **kwargs):
        print("activate")

    @Gtk.Template.Callback()
    def on_search_entry_stop_cb(self, *arg, **kwargs):
        print("stop")

    def on_delete_click(self, _widget, item):
        self.recents_manager.remove_item(item.uri)


# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from gi.repository import Gio


class Settings(Gio.Settings):

    """
        Apostrophe Settings
    """

    def __init__(self):
        """
            Init Settings
        """
        Gio.Settings.__init__(self)

    @classmethod
    def new(cls):
        """
            Return a new Settings object
        """
        settings = Gio.Settings.new("org.gnome.gitlab.somas.Apostrophe")
        settings.__class__ = Settings
        return settings



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

class TextViewScroller:
    def __init__(self, text_view, scrolled_window):
        super().__init__()

        self.text_view = text_view
        self.scrolled_window = scrolled_window
        self.smooth_scroller = None

    def can_scroll(self):
        vap = self.scrolled_window.get_vadjustment().props
        return vap.upper > vap.page_size

    def get_scroll_scale(self):
        vap = self.scrolled_window.get_vadjustment().props
        if vap.upper > vap.page_size:
            return vap.value / (vap.upper - vap.page_size)
        else:
            return 0

    def set_scroll_scale(self, scale):
        vap = self.scrolled_window.get_vadjustment().props
        vap.value = (vap.upper - vap.page_size) * scale

    def scroll_to_mark(self, mark, center):
        """Scrolls until mark is visible, if needed."""

        target_pos = self.get_target_pos_for_mark(mark, center)
        if target_pos:
            self.scrolled_window.get_vadjustment().set_value(target_pos)

    def smooth_scroll_to_mark(self, mark, center):
        """Smoothly scrolls until mark is visible, if needed."""

        if self.smooth_scroller and self.smooth_scroller.is_started:
            self.smooth_scroller.end()

        target_pos = self.get_target_pos_for_mark(mark, center)
        if target_pos:
            source_pos = self.scrolled_window.get_vadjustment().props.value
            self.smooth_scroller = SmoothScroller(
                self.scrolled_window, source_pos, target_pos)
            self.smooth_scroller.start()

    def get_target_pos_for_mark(self, mark, center):
        margin = 32

        mark_iter = self.text_view.get_buffer().get_iter_at_mark(mark)
        mark_rect = self.text_view.get_iter_location(mark_iter)

        vap = self.scrolled_window.get_vadjustment().props

        pos_y = mark_rect.y + mark_rect.height + self.text_view.props.top_margin
        pos_viewport_y = pos_y - vap.value
        target_pos = None
        if center:
            if pos_viewport_y != vap.page_size / 2:
                target_pos = pos_y - (vap.page_size / 2)
        elif pos_viewport_y > vap.page_size - margin:
            target_pos = pos_y - vap.page_size + margin
        elif pos_viewport_y < margin:
            target_pos = pos_y - margin - mark_rect.height

        return target_pos


class SmoothScroller:
    def __init__(self, scrolled_window, source_pos, target_pos):
        super().__init__()

        self.scrolled_window = scrolled_window
        self.source_pos = source_pos
        self.target_pos = target_pos
        self.duration = max(100, (target_pos - source_pos) / 50) * 4000

        self.is_started = False
        self.is_setup = False
        self.start_time = 0
        self.end_time = 0
        self.tick_callback_id = 0

    def start(self):
        self.is_started = True
        self.tick_callback_id = self.scrolled_window.add_tick_callback(
            self.on_tick)

    def end(self):
        self.scrolled_window.remove_tick_callback(self.tick_callback_id)
        self.is_started = False

    def setup(self, time):
        self.start_time = time
        self.end_time = time + self.duration
        self.is_setup = True

    def on_tick(self, widget, frame_clock):
        def ease_out_cubic(value):
            return pow(value - 1, 3) + 1

        now = frame_clock.get_frame_time()
        if not self.is_setup:
            self.setup(now)

        if now < self.end_time:
            time = float(now - self.start_time) / \
                float(self.end_time - self.start_time)
        else:
            time = 1
            self.end()

        time = ease_out_cubic(time)
        pos = self.source_pos + (time * (self.target_pos - self.source_pos))
        widget.get_vadjustment().props.value = pos
        return True



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import logging
import os
from gettext import gettext as _

from dataclasses import dataclass
import chardet

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, GLib, Gio, Handy

from apostrophe.export_dialog import ExportDialog, AdvancedExportDialog
from apostrophe.preview_handler import PreviewHandler
from apostrophe.stats_handler import StatsHandler
from apostrophe.text_view import ApostropheTextView
from apostrophe.search_and_replace import ApostropheSearchBar
from apostrophe.settings import Settings
from apostrophe.tweener import Tweener
from apostrophe.helpers import App
from apostrophe import helpers

from .headerbars import BaseHeaderbar

LOGGER = logging.getLogger('apostrophe')


@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/Window.ui')
class MainWindow(Handy.ApplicationWindow):

    __gtype_name__ = "ApostropheWindow"


    editor_scrolledwindow = Gtk.Template.Child()
    save_progressbar = Gtk.Template.Child()
    headerbar = Gtk.Template.Child()
    headerbar_eventbox = Gtk.Template.Child()
    searchbar = Gtk.Template.Child()
    stats_revealer = Gtk.Template.Child()
    stats_button = Gtk.Template.Child()
    flap = Gtk.Template.Child()
    # TODO ??
    preview_stack = Gtk.Template.Child()
    text_view = Gtk.Template.Child()
    editor = Gtk.Template.Child()

    subtitle = GObject.Property(type=str)
    is_fullscreen = GObject.Property(type=bool, default=False)
    headerbar_visible = GObject.Property(type=bool, default=True)
    bottombar_visible = GObject.Property(type=bool, default=True)

    preview = GObject.Property(type=bool, default=False)
    preview_layout = GObject.Property(type=int, default=1)

    did_change = GObject.Property(type=bool, default=False)

    def __init__(self, app):
        """Set up the main window"""

        super().__init__(application=Gio.Application.get_default(),
                         title="Apostrophe")

        #TODO: size

        # Preferences
        self.settings = Settings.new()

        # Create new, empty file
        # TODO: load last opened file?

        self.current = File()

        # Setup text editor
        self.text_view.get_buffer().connect('changed', self.on_text_changed)
        self.text_view.grab_focus()

        # Setup save progressbar an its animator
        self.progressbar_initiate_tw = Tweener(self.save_progressbar,
                                               self.save_progressbar.set_fraction,
                                               0, 0.125, 40)
        self.progressbar_finalize_tw = Tweener(self.save_progressbar,
                                               self.save_progressbar.set_fraction,
                                               0.125, 1, 400)
        self.progressbar_opacity_tw = Tweener(self.save_progressbar,
                                              self.save_progressbar.set_opacity,
                                              1, 0, 300, 200,
                                              callback = self.save_progressbar.set_visible,
                                              callback_arg = False)

        # Setup stats counter
        self.stats_handler = StatsHandler(self.stats_button, self.text_view)

        # Setup preview
        self.preview_handler = PreviewHandler(self, self.text_view, self.flap)

        # Setting up spellcheck
        self.settings.bind("spellcheck", self.text_view,
                           "spellcheck", Gio.SettingsBindFlags.GET)

        # Setting up text size
        self.settings.bind("bigger-text", self.text_view,
                           "bigger_text", Gio.SettingsBindFlags.GET)

        self.settings.bind("characters-per-line", self.text_view,
                           "line_chars", Gio.SettingsBindFlags.GET)

        # Search and replace initialization
        self.searchbar.attach(self.text_view)

        # Actions
        action = Gio.PropertyAction.new("find", self.searchbar, "search-mode-enabled")
        self.add_action(action)

        action = Gio.PropertyAction.new("find_replace", self.searchbar, "replace-mode-enabled")
        self.add_action(action)

        action = Gio.PropertyAction.new("focus_mode", self.text_view, "focus-mode")
        self.add_action(action)

        action = Gio.PropertyAction.new("hemingway_mode", self.text_view, "hemingway-mode")
        self.add_action(action)

        action = Gio.PropertyAction.new("fullscreen", self, "is-fullscreen")
        self.add_action(action)

        action = Gio.PropertyAction.new("preview", self, "preview")
        self.add_action(action)
        self.connect("notify::preview", self.toggle_preview)


        action = Gio.SimpleAction.new("new", None)
        action.connect("activate", self.new_document)
        self.add_action(action)

        action = Gio.SimpleAction.new("open", None)
        action.connect("activate", self.open_document)
        self.add_action(action)

        action = Gio.SimpleAction.new("open_file", GLib.VariantType("s"))
        action.connect("activate", self.open_from_gvariant)
        self.add_action(action)

        action = Gio.SimpleAction.new("save", None)
        action.connect("activate", self.save_document)
        self.add_action(action)

        action = Gio.SimpleAction.new("save_as", None)
        action.connect("activate", self.save_document_as)
        self.add_action(action)

        action = Gio.SimpleAction.new("export", GLib.VariantType("s"))
        action.connect("activate", self.open_export)
        self.add_action(action)

        action = Gio.SimpleAction.new("advanced_export", None)
        action.connect("activate", self.open_advanced_export)
        self.add_action(action)

        action = Gio.SimpleAction.new("copy_html", None)
        action.connect("activate", self.copy_html_to_clipboard)
        self.add_action(action)

        # not really necessary but we'll keep a preview_layout property on the window
        # and bind it both to the switcher and the renderer
        self.preview_layout = self.settings.get_enum("preview-mode")
        self.bind_property("preview_layout", self.headerbar.preview_layout_switcher, 
                           "preview_layout", GObject.BindingFlags.BIDIRECTIONAL|GObject.BindingFlags.SYNC_CREATE)

        self.bind_property("preview_layout", self.preview_handler.preview_renderer, 
                           "preview_layout", GObject.BindingFlags.SYNC_CREATE)

        self.preview = self.settings.get_boolean("preview-active")

        self.text_view.hemingway_mode = self.settings.get_boolean("hemingway-mode")

        self.new_document()

    # TODO: change to closures on GTK4
    # "Bind" scrolled window's scrollbar margin to headerbar visibility
    @Gtk.Template.Callback()
    def headerbar_revealed_cb(self, widget, *args):
        scrollbar = self.editor_scrolledwindow.get_vscrollbar()
        scrollbar.set_margin_top(46 if widget.get_reveal_child() else 0)

    def on_text_changed(self, *_args):
        """called when the text changes, sets the self.did_change to true and
           updates the title and the counters to reflect that
        """

        if self.did_change is False:
            self.did_change = True
        self.update_headerbar_title(True, True)
        if self.settings.get_value("autohide-headerbar"):
            self.hide_headerbar_bottombar()

    @Gtk.Template.Callback()
    def _on_fullscreen(self, *args, **kwargs):
        """Puts the application in fullscreen mode and show/hides
        the poller for motion in the top border
        """
        if self.is_fullscreen == True:
            self.fullscreen()
            self.hide_headerbar_bottombar()
        else:
            self.unfullscreen()
            self.reveal_headerbar_bottombar()

        self.text_view.grab_focus()

    def toggle_preview(self, *args, **kwargs):
        """Toggle the preview mode
        """

        if self.preview:
            self.text_view.grab_focus()
            self.preview_handler.show()
        else:
            self.preview_handler.hide()
            self.text_view.grab_focus()

    def save_document(self, _action=None, _value=None, sync: bool = False) -> bool:
        """Try to save buffer in the current gfile.
        If the file doesn't exist calls save_document_as

        Args:
            sync (bool, optional): Wheter the save operation should be done
            synchronously. Defaults to False.

        Returns:
            bool: True if the document was saved correctly
        """

        if self.current.gfile:
            LOGGER.info("saving")

            # We try to encode the file with the given encoding
            # if that doesn't work, we try with UTF-8
            # if that fails as well, we return False
            try:
                try:
                    encoded_text = self.text_view.get_text()\
                        .encode(self.current.encoding)
                except UnicodeEncodeError:
                    encoded_text = self.text_view.get_text()\
                        .encode("UTF-8")
                    self.current.encoding = "UTF-8"
            except UnicodeEncodeError as error:
                helpers.show_error(self, str(error.reason))
                LOGGER.warning(str(error.reason))
                return False
            else:
                self.save_progressbar.set_opacity(1)
                self.save_progressbar.set_visible(True)
                self.progressbar_initiate_tw.start()

                # we allow synchronously saving operations
                # for result-dependant code
                if sync:
                    try:
                        res = self.current.gfile.replace_contents(
                            encoded_text,
                            etag=None,
                            make_backup=False,
                            flags=Gio.FileCreateFlags.NONE,
                            cancellable=None)
                    except GLib.GError as error:
                        helpers.show_error(self, str(error.message))
                        LOGGER.warning(str(error.message))
                        self.progressbar_opacity_tw.start()
                        self.did_change = True
                        return False

                    if res:
                        self.progressbar_initiate_tw.stop()
                        self.progressbar_finalize_tw.start()
                        self.progressbar_opacity_tw.start()

                        self.update_headerbar_title()
                        self.did_change = False
                        return True

                    else:
                        self.progressbar_opacity_tw.start()
                        self.did_change = True
                        return False

                else:
                    self.current.gfile.replace_contents_bytes_async(
                        GLib.Bytes.new(encoded_text),
                        etag=None,
                        make_backup=False,
                        flags=Gio.FileCreateFlags.NONE,
                        cancellable=None,
                        callback=self._replace_contents_cb,
                        user_data=None)
                    return True
        # if there's no GFile we ask for one:
        else:
            return self.save_document_as(sync=sync)

    def save_document_as(self, _widget=None, _data=None,
                         sync: bool = False) -> bool:
        """provide to the user a filechooser and save the document
           where they want. Call set_headbar_title after that
        """

        filefilter = Gtk.FileFilter.new()
        filefilter.add_mime_type('text/x-markdown')
        filefilter.add_mime_type('text/plain')
        filefilter.set_name('Markdown (.md)')
        filechooser = Gtk.FileChooserNative.new(
            _("Save your File"),
            self,
            Gtk.FileChooserAction.SAVE,
            _("Save"),
            _("Cancel")
        )
        filechooser.set_do_overwrite_confirmation(True)
        filechooser.set_local_only(False)
        filechooser.add_filter(filefilter)
        filechooser.set_modal(True)
        filechooser.set_transient_for(self)

        title = self.current.title
        if not title.endswith(".md"):
            title += ".md"
        filechooser.set_current_name(title)

        response = filechooser.run()

        if response == Gtk.ResponseType.ACCEPT:

            file = filechooser.get_file()

            if not file.query_exists():
                try:
                    file.create(Gio.FileCreateFlags.NONE)
                except GLib.GError as error:
                    helpers.show_error(self, str(error.message))
                    LOGGER.warning(str(error.message))
                    return False

            self.current.gfile = file

            recents_manager = Gtk.RecentManager.get_default()
            recents_manager.add_item(self.current.gfile.get_uri())

            self.update_headerbar_title(False, True)
            filechooser.destroy()
            return self.save_document()
        else:
            return False

    def _replace_contents_cb(self, gfile, result, _user_data=None):
        try:
            success, _etag = gfile.replace_contents_finish(result)
        except GLib.GError as error:
            helpers.show_error(self, str(error.message))
            LOGGER.warning(str(error.message))
            self.progressbar_opacity_tw.start()
            self.did_change = True
            return False

        if success:
            self.progressbar_initiate_tw.stop()
            self.progressbar_finalize_tw.start()
            self.progressbar_opacity_tw.start()

            self.update_headerbar_title()
            self.did_change = False
        else:
            self.did_change = True
            self.progressbar_opacity_tw.start()

        return success

    def copy_html_to_clipboard(self, _widget=None, _date=None):
        """Copies only html without headers etc. to Clipboard
        """

        output = helpers.pandoc_convert(self.text_view.get_text())
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(output, -1)
        clipboard.store()

    def open_document(self, _action, _value):
        """open the desired file
        """

        response = self.check_change()
        if (response == Gtk.ResponseType.CANCEL or
            response == Gtk.ResponseType.DELETE_EVENT):
            return

        markdown_filter = Gtk.FileFilter.new()
        markdown_filter.add_mime_type('text/markdown')
        markdown_filter.add_mime_type('text/x-markdown')
        markdown_filter.set_name(_('Markdown Files'))

        plaintext_filter = Gtk.FileFilter.new()
        plaintext_filter.add_mime_type('text/plain')
        plaintext_filter.set_name(_('Plain Text Files'))

        filechooser = Gtk.FileChooserNative.new(
            _("Open a .md file"),
            self,
            Gtk.FileChooserAction.OPEN,
            _("Open"),
            _("Cancel")
        )

        filechooser.set_local_only(False)
        filechooser.add_filter(markdown_filter)
        filechooser.add_filter(plaintext_filter)
        response = filechooser.run()
        if response == Gtk.ResponseType.ACCEPT:
            self.load_file(filechooser.get_file())
            filechooser.destroy()

        elif response == Gtk.ResponseType.CANCEL:
            filechooser.destroy()

    def open_from_gvariant(self, _action, gvariant):
        self.load_file(Gio.File.new_for_uri(gvariant.get_string()))

    def load_file(self, file=None):
        """Open File from command line or open / open recent etc."""
        LOGGER.info("trying to open %s", file.get_path())

        response = self.check_change()
        if (response == Gtk.ResponseType.CANCEL or
            response == Gtk.ResponseType.DELETE_EVENT):
            return
        self.current.gfile = file

        self.current.gfile.load_contents_async(None,
                                               self._load_contents_cb, None)

    def _load_contents_cb(self, gfile, result, user_data=None):
        try:
            _success, contents, _etag = gfile.load_contents_finish(result)
        except GLib.GError as error:
            helpers.show_error(self, str(error.message))
            LOGGER.warning(str(error.message))
            return

        try:
            try:
                self.current.encoding = 'UTF-8'
                decoded = contents.decode(self.current.encoding)
            except UnicodeDecodeError:
                self.current.encoding = chardet.detect(contents)['encoding']
                decoded = contents.decode(self.current.encoding)
        except UnicodeDecodeError as error:
            helpers.show_error(self, str(error.message))
            LOGGER.warning(str(error.message))
            return
        else:
            self.text_view.set_text(decoded)
            start_iter = self.text_view.get_buffer().get_start_iter()
            GLib.idle_add(
                lambda: self.text_view.get_buffer().place_cursor(start_iter))

            ## add file to recents manager once it's fully loaded,
            # unless it is an internal resource
            if not self.current.gfile.get_uri().startswith("resource:"):
                recents_manager = Gtk.RecentManager.get_default()
                recents_manager.add_item(self.current.gfile.get_uri())

            self.update_headerbar_title()

            self.did_change = False

    def check_change(self) -> Gtk.ResponseType:
        """Show dialog to prevent loss of unsaved changes
        """

        if self.did_change and self.text_view.get_text():
            dialog = Gtk.MessageDialog(self,
                                       Gtk.DialogFlags.MODAL |
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                       Gtk.MessageType.WARNING,
                                       Gtk.ButtonsType.NONE,
                                       _("Save changes to document " +
                                         "“%s” before closing?") %
                                       self.current.title
                                       )

            dialog.props.secondary_text = _("If you don’t save, " +
                                            "all your changes will be " +
                                            "permanently lost.")
            close_button = dialog.add_button(_("Close without saving"),
                                             Gtk.ResponseType.NO)
            dialog.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
            dialog.add_button(_("Save now"), Gtk.ResponseType.YES)

            close_button.get_style_context().add_class("destructive-action")
            dialog.set_default_response(Gtk.ResponseType.YES)
            response = dialog.run()

            dialog.destroy()

            if response == Gtk.ResponseType.YES:
                # If the saving fails, retry
                if self.save_document(sync=True) is False:
                    return self.check_change()

            return response

    def new_document(self, *args, **kwargs):
        """create new document
        """

        response = self.check_change()
        if (response == Gtk.ResponseType.CANCEL or
            response == Gtk.ResponseType.DELETE_EVENT):
            return
        self.text_view.clear()

        self.did_change = False
        self.current.gfile = None
        self.update_headerbar_title(False, False)

    def update_default_stat(self):
        self.stats_handler.update_default_stat()

    def reload_preview(self, reshow=False):
        self.preview_handler.reload(reshow=reshow)

    def open_export(self, _action, value):
        """open the export dialog
        """
        text = bytes(self.text_view.get_text(), "utf-8")

        export_format = value.get_string()

        export_dialog = ExportDialog(self.current, export_format, text)
        export_dialog.dialog.set_transient_for(self)
        export_dialog.export()

    def open_advanced_export(self, *args, **kwargs):
        """open the advanced export dialog
        """
        text = bytes(self.text_view.get_text(), "utf-8")

        export_dialog = AdvancedExportDialog(self.current, text)
        export_dialog.set_transient_for(self)
        export_dialog.show()

    @Gtk.Template.Callback()
    def reveal_headerbar_bottombar(self, _widget=None, _data=None):

        self.reveal_bottombar()

        if not self.headerbar_visible:
            self.headerbar_eventbox.hide()
            self.headerbar.set_visible(True)
            self.headerbar.set_reveal_child(True)
            self.get_style_context().remove_class("focus")
            self.headerbar_visible = True

    @Gtk.Template.Callback()
    def reveal_bottombar(self, _widget=None, _data=None):

        if not self.bottombar_visible:
            self.stats_revealer.set_reveal_child(True)
            self.stats_revealer.set_halign(Gtk.Align.END)
            self.stats_revealer.queue_resize()

            self.bottombar_visible = True

    def hide_headerbar_bottombar(self):

        if self.searchbar.get_search_mode():
            return

        if self.headerbar_visible:
            self.headerbar.set_reveal_child(False)
            self.get_style_context().add_class("focus")

            self.headerbar_visible = False

        if self.bottombar_visible:
            self.stats_revealer.set_reveal_child(False)
            self.stats_revealer.set_halign(Gtk.Align.FILL)

            self.bottombar_visible = False

        self.headerbar_eventbox.show()

    # TODO: this has to go
    def update_headerbar_title(self,
                               is_unsaved: bool = False,
                               has_subtitle: bool = True):
        """update headerbar title and subtitle
        """

        if is_unsaved:
            prefix = "• "
            # TODO: this doesn't really belong here
            App().inhibitor.inhibit(Gtk.ApplicationInhibitFlags.LOGOUT)
        else:
            prefix = ""
            App().inhibitor.uninhibit()

        title = prefix + self.current.title

        if has_subtitle:
            subtitle = self.current.path
        else:
            subtitle = ""

        self.set_title(title)
        self.subtitle = subtitle
        self.headerbar.set_tooltip_text(subtitle)

    def save_state(self):
        self.settings.set_enum("preview-mode", self.preview_layout)
        self.settings.set_boolean("preview-active", self.preview)
        self.settings.set_boolean("hemingway-mode", self.text_view.hemingway_mode)

    @Gtk.Template.Callback()
    def on_delete_called(self, _widget, _data=None):
        """Called when the ApostropheWindow is closed.
        """
        LOGGER.info('delete called')
        response = self.check_change()
        if (response == Gtk.ResponseType.CANCEL or
            response == Gtk.ResponseType.DELETE_EVENT):
            return True

        # save state if we're the last window OR if only the preview window is left
        n_windows = len(self.get_application().get_windows())
        if n_windows == 1 or \
           (n_windows == 2 and self.preview_handler.preview_renderer.window):
            self.save_state()
        return False

@dataclass
class File():
    """Class for keeping track of files, their attributes, and their methods"""

    def __init__(self, gfile=None, encoding="UTF-8"):
        self._settings = Settings.new()
        self.gfile = gfile
        self.encoding = encoding
        self.path = ""
        self.title = ""
        self.name = ""

    @property
    def gfile(self):
        return self._gfile

    @gfile.setter
    def gfile(self, file):
        if file:
            if file.is_native():
                self.path = file.get_parent().get_path()
                base_path = file.get_parent().get_path()
                os.chdir(base_path)
            else:
                self.path = file.get_parent().get_uri()
                base_path = "/"

            file_info = file.query_info("standard",
                                        Gio.FileQueryInfoFlags.NONE,
                                        None)
            self.title = file_info.get_attribute_as_string(
                "standard::display-name")
        else:
            self.title = _("New File")
            base_path = "/"
        self.name = self.title
        if self.name.endswith(".md"):
            self.name = self.name[:-3]
        # TODO: remove path in favor of gfile
        self._settings.set_string("open-file-path", base_path)
        self._gfile = file
